import numpy as np
import scipy.sparse as sp
import SimPEG
from .. import Utils
from SimPEG.EM.Utils import omega
from SimPEG.Utils import Zero, Identity, sdiag


class FieldsFDEM(SimPEG.Problem.Fields):
    """

    Fancy Field Storage for a FDEM survey. Only one field type is stored for
    each problem, the rest are computed. The fields object acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[srcList,'e']
        b = f[srcList,'b']

    If accessing all sources for a given field, use the :code:`:`

    .. code-block:: python

        f = problem.fields(m)
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (nE or nF, nSrcs :math:`\\times`
    nFrequencies)
    """

    knownFields = {}
    dtype = complex

    def _GLoc(self, fieldType):
        """Grid location of the fieldType"""
        return self.aliasFields[fieldType][1]

    def _e(self, solution, srcList):
        """
        Total electric field is sum of primary and secondary

        :param numpy.ndarray solution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total electric field
        """
        if (
            getattr(self, "_ePrimary", None) is None
            or getattr(self, "_eSecondary", None) is None
        ):
            raise NotImplementedError(
                "Getting e from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        return self._ePrimary(solution, srcList) + self._eSecondary(solution, srcList)

    def _b(self, solution, srcList):
        """
        Total magnetic flux density is sum of primary and secondary

        :param numpy.ndarray solution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total magnetic flux density
        """
        if (
            getattr(self, "_bPrimary", None) is None
            or getattr(self, "_bSecondary", None) is None
        ):
            raise NotImplementedError(
                "Getting b from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        return self._bPrimary(solution, srcList) + self._bSecondary(solution, srcList)

    def _bSecondary(self, solution, srcList):
        """
        Total magnetic flux density is sum of primary and secondary

        :param numpy.ndarray solution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total magnetic flux density
        """
        if getattr(self, "_bSecondary", None) is None:
            raise NotImplementedError(
                "Getting b from {} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        return self._bSecondary(solution, srcList)

    def _h(self, solution, srcList):
        """
        Total magnetic field is sum of primary and secondary

        :param numpy.ndarray solution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total magnetic field
        """
        if (
            getattr(self, "_hPrimary", None) is None
            or getattr(self, "_hSecondary", None) is None
        ):
            raise NotImplementedError(
                "Getting h from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        return self._hPrimary(solution, srcList) + self._hSecondary(solution, srcList)

    def _j(self, solution, srcList):
        """
        Total current density is sum of primary and secondary

        :param numpy.ndarray solution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total current density
        """
        if (
            getattr(self, "_jPrimary", None) is None
            or getattr(self, "_jSecondary", None) is None
        ):
            raise NotImplementedError(
                "Getting j from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        return self._jPrimary(solution, srcList) + self._jSecondary(solution, srcList)

    def _eDeriv(self, src, du_dm_v, v, adjoint=False):
        r"""
        Total derivative of e with respect to the inversion model. Returns
        :math:`d\mathbf{e}/d\mathbf{m}` for forward and
        (:math:`d\mathbf{e}/d\mathbf{u}`, :math:`d\mathb{u}/d\mathbf{m}`)
        for the adjoint

        :param SimPEG.EM.FDEM.Src.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: derivative of the solution vector with
            respect to the model times a vector (is None for adjoint)
        :param numpy.ndarray v: vector to take sensitivity product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: derivative times a vector (or tuple for adjoint)
        """
        if (
            getattr(self, "_eDeriv_u", None) is None
            or getattr(self, "_eDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting eDerivs from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        if adjoint:
            return (self._eDeriv_u(src, v, adjoint), self._eDeriv_m(src, v, adjoint))
        return np.array(
            self._eDeriv_u(src, du_dm_v, adjoint) + self._eDeriv_m(src, v, adjoint),
            dtype=complex,
        )

    def _bDeriv(self, src, du_dm_v, v, adjoint=False):
        r"""
        Total derivative of b with respect to the inversion model. Returns
        :math:`d\mathbf{b}/d\mathbf{m}` for forward and
        (:math:`d\mathbf{b}/d\mathbf{u}`, :math:`d\mathb{u}/d\mathbf{m}`) for
        the adjoint

        :param SimPEG.EM.FDEM.Src.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: derivative of the solution vector with
            respect to the model times a vector (is None for adjoint)
        :param numpy.ndarray v: vector to take sensitivity product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: derivative times a vector (or tuple for adjoint)
        """
        if (
            getattr(self, "_bDeriv_u", None) is None
            or getattr(self, "_bDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting bDerivs from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        if adjoint:
            return (self._bDeriv_u(src, v, adjoint), self._bDeriv_m(src, v, adjoint))
        return np.array(
            self._bDeriv_u(src, du_dm_v, adjoint) + self._bDeriv_m(src, v, adjoint),
            dtype=complex,
        )

    def _bSecondaryDeriv(self, src, du_dm_v, v, adjoint=False):
        r"""
        Total derivative of b with respect to the inversion model. Returns
        :math:`d\mathbf{b}/d\mathbf{m}` for forward and
        (:math:`d\mathbf{b}/d\mathbf{u}`, :math:`d\mathb{u}/d\mathbf{m}`) for
        the adjoint

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: sorce
        :param numpy.ndarray du_dm_v: derivative of the solution vector with
            respect to the model times a vector (is None for adjoint)
        :param numpy.ndarray v: vector to take sensitivity product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: derivative times a vector (or tuple for adjoint)
        """
        # TODO: modify when primary field is dependent on m

        return self._bDeriv(src, du_dm_v, v, adjoint=adjoint)

    def _hDeriv(self, src, du_dm_v, v, adjoint=False):
        r"""
        Total derivative of h with respect to the inversion model. Returns
        :math:`d\mathbf{h}/d\mathbf{m}` for forward and
        (:math:`d\mathbf{h}/d\mathbf{u}`, :math:`d\mathb{u}/d\mathbf{m}`)
        for the adjoint

        :param SimPEG.EM.FDEM.Src.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: derivative of the solution vector with
            respect to the model times a vector (is None for adjoint)
        :param numpy.ndarray v: vector to take sensitivity product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: derivative times a vector (or tuple for adjoint)
        """
        if (
            getattr(self, "_hDeriv_u", None) is None
            or getattr(self, "_hDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting hDerivs from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        if adjoint:
            return (self._hDeriv_u(src, v, adjoint), self._hDeriv_m(src, v, adjoint))
        return np.array(
            self._hDeriv_u(src, du_dm_v, adjoint) + self._hDeriv_m(src, v, adjoint),
            dtype=complex,
        )

    def _jDeriv(self, src, du_dm_v, v, adjoint=False):
        r"""
        Total derivative of j with respect to the inversion model. Returns
        :math:`d\mathbf{j}/d\mathbf{m}` for forward and
        (:math:`d\mathbf{j}/d\mathbf{u}`, :math:`d\mathb{u}/d\mathbf{m}`) for
        the adjoint

        :param SimPEG.EM.FDEM.Src.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: derivative of the solution vector with
            respect to the model times a vector (is None for adjoint)
        :param numpy.ndarray v: vector to take sensitivity product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: derivative times a vector (or tuple for adjoint)
        """
        if (
            getattr(self, "_jDeriv_u", None) is None
            or getattr(self, "_jDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting jDerivs from {!s} is not implemented".format(
                    self.knownFields.keys()[0]
                )
            )

        if adjoint:
            return (self._jDeriv_u(src, v, adjoint), self._jDeriv_m(src, v, adjoint))
        return np.array(
            self._jDeriv_u(src, du_dm_v, adjoint) + self._jDeriv_m(src, v, adjoint),
            dtype=complex,
        )


class Fields3D_e(FieldsFDEM):
    """
    Fields object for Problem3D_e.

    :param discretize.BaseMesh.BaseMesh mesh: mesh
    :param SimPEG.EM.FDEM.SurveyFDEM.Survey survey: survey
    """

    knownFields = {"eSolution": "E"}
    aliasFields = {
        "e": ["eSolution", "E", "_e"],
        "ePrimary": ["eSolution", "E", "_ePrimary"],
        "eSecondary": ["eSolution", "E", "_eSecondary"],
        "b": ["eSolution", "F", "_b"],
        "bPrimary": ["eSolution", "F", "_bPrimary"],
        "bSecondary": ["eSolution", "F", "_bSecondary"],
        "j": ["eSolution", "CCV", "_j"],
        "h": ["eSolution", "CCV", "_h"],
    }

    def startup(self):
        self.prob = self.survey.prob
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._aveE2CCV = self.survey.prob.mesh.aveE2CCV
        self._aveF2CCV = self.survey.prob.mesh.aveF2CCV
        self._nC = self.survey.prob.mesh.nC
        self._MeSigma = self.survey.prob.MeSigma
        self._MeSigmaDeriv = self.survey.prob.MeSigmaDeriv
        self._MfMui = self.survey.prob.MfMui
        self._MfMuiDeriv = self.survey.prob.MfMuiDeriv

    def _GLoc(self, fieldType):
        if fieldType in ["e", "eSecondary", "ePrimary"]:
            return "E"
        elif fieldType in ["b", "bSecondary", "bPrimary"]:
            return "F"
        elif (fieldType == "h") or (fieldType == "j"):
            return "CCV"
        else:
            raise Exception("Field type must be e, b, h, j")

    def _ePrimary(self, eSolution, srcList):
        """
        Primary electric field from source

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary electric field as defined by the sources
        """

        ePrimary = np.zeros([self.prob.mesh.nE, len(srcList)], dtype=complex)
        for i, src in enumerate(srcList):
            ep = src.ePrimary(self.prob)
            ePrimary[:, i] = ePrimary[:, i] + ep
        return ePrimary

    def _eSecondary(self, eSolution, srcList):
        """
        Secondary electric field is the thing we solved for

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary electric field
        """
        return eSolution

    def _eDeriv_u(self, src, v, adjoint=False):
        """
        Partial derivative of the total electric field with respect to the
        thing we solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the field we solved for with a vector
        """

        return Identity() * v

    def _eDeriv_m(self, src, v, adjoint=False):
        """
        Partial derivative of the total electric field with respect to the
        inversion model. Here, we assume that the primary does not depend on
        the model. Note that this also includes derivative contributions from
        the sources.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: SimPEG.Utils.Zero
        :return: product of the electric field derivative with respect to the
            inversion model with a vector
        """

        return src.ePrimaryDeriv(self.prob, v, adjoint)

    def _bPrimary(self, eSolution, srcList):
        """
        Primary magnetic flux density from source

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary magnetic flux density as defined by the sources
        """

        bPrimary = np.zeros(
            [self._edgeCurl.shape[0], eSolution.shape[1]], dtype=complex
        )

        for i, src in enumerate(srcList):
            bp = src.bPrimary(self.prob)
            bPrimary[:, i] = bPrimary[:, i] + bp
        return bPrimary

    def _bSecondary(self, eSolution, srcList):
        """
        Secondary magnetic flux density from eSolution

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary magnetic flux density
        """

        C = self._edgeCurl
        b = C * eSolution
        for i, src in enumerate(srcList):
            b[:, i] *= -1.0 / (1j * omega(src.freq))  # freq depends on the source
            s_m = src.s_m(self.prob)
            b[:, i] = b[:, i] + 1.0 / (1j * omega(src.freq)) * s_m
        return b

    def _bDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the thing we
        solved for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic flux density with
            respect to the field we solved for with a vector
        """

        C = self._edgeCurl
        if adjoint:
            return -1.0 / (1j * omega(src.freq)) * (C.T * du_dm_v)
        return -1.0 / (1j * omega(src.freq)) * (C * du_dm_v)

    def _bDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the inversion
        model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the magnetic flux density derivative with respect
            to the inversion model with a vector
        """

        return self._bDeriv_src(src, v, adjoint=adjoint)

    def _bDeriv_src(self, src, v, adjoint=False):
        s_mDeriv = src.s_mDeriv(self.prob, v, adjoint)
        return 1.0 / (1j * omega(src.freq)) * s_mDeriv + src.bPrimaryDeriv(
            self.prob, v, adjoint
        )

    def _j(self, eSolution, srcList):
        """
        Current density from eSolution

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: current density
        """
        aveE2CCV = self._aveE2CCV
        # number of components (instead of checking if cyl or not)
        n = int(aveE2CCV.shape[0] / self._nC)
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        return VI * (aveE2CCV * (self._MeSigma * self._e(eSolution, srcList)))

    def _jDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the current density with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the current density with respect
            to the field we solved for with a vector
        """
        # number of components (instead of checking if cyl or not)
        n = int(self._aveE2CCV.shape[0] / self._nC)
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint:
            return self._eDeriv_u(
                src,
                self._MeSigma.T * (self._aveE2CCV.T * (VI.T * du_dm_v)),
                adjoint=adjoint,
            )
        return VI * (
            self._aveE2CCV
            * (self._MeSigma * (self._eDeriv_u(src, du_dm_v, adjoint=adjoint)))
        )

    def _jDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the current density with respect to the inversion model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the current density derivative with respect to the
            inversion model with a vector
        """
        e = self[src, "e"]
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint:
            return (
                self._MeSigmaDeriv(e).T * (self._aveE2CCV.T * (VI.T * v))
                + self._eDeriv_m(src, self._aveE2CCV.T * (VI.T * v), adjoint=adjoint)
            ) + src.jPrimaryDeriv(self.prob, v, adjoint)
        return (
            VI
            * (
                self._aveE2CCV
                * (self._eDeriv_m(src, v, adjoint=adjoint) + self._MeSigmaDeriv(e) * v)
            )
        ) + src.jPrimaryDeriv(self.prob, v, adjoint)

    def _h(self, eSolution, srcList):
        """
        Magnetic field from eSolution

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: magnetic field
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # Number of Components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        return VI * (self._aveF2CCV * (self._MfMui * self._b(eSolution, srcList)))

    def _hDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the magnetic field with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the field we solved for with a vector
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # Number of Components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            v = self._MfMui.T * (self._aveF2CCV.T * (VI.T * du_dm_v))
            return self._bDeriv_u(src, v, adjoint=adjoint)
        return VI * (
            self._aveF2CCV
            * (self._MfMui * self._bDeriv_u(src, du_dm_v, adjoint=adjoint))
        )

    def _hDeriv_mui(self, src, v, adjoint=False):
        n = int(self._aveF2CCV.shape[0] / self._nC)  # Number of Components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint is True:
            return self._MfMuiDeriv(self[src, "b"]).T * (self._aveF2CCV.T * (VI.T * v))

        return VI * (self._aveF2CCV * (self._MfMuiDeriv(self[src, "b"]) * v))

    def _hDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic field with respect to the inversion model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the magnetic field derivative with respect to the
            inversion model with a vector
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # Number of Components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._bDeriv_m(
                src, self._MfMui.T * (self._aveF2CCV.T * (VI.T * v)), adjoint=adjoint
            ) + self._hDeriv_mui(src, v, adjoint=adjoint)
        return (
            VI
            * (self._aveF2CCV * (self._MfMui * self._bDeriv_m(src, v, adjoint=adjoint)))
        ) + self._hDeriv_mui(src, v, adjoint=adjoint)


class Fields3D_b(FieldsFDEM):
    """
    Fields object for Problem3D_b.

    :param discretize.BaseMesh.BaseMesh mesh: mesh
    :param SimPEG.EM.FDEM.SurveyFDEM.Survey survey: survey
    """

    knownFields = {"bSolution": "F"}
    aliasFields = {
        "b": ["bSolution", "F", "_b"],
        "bPrimary": ["bSolution", "F", "_bPrimary"],
        "bSecondary": ["bSolution", "F", "_bSecondary"],
        "e": ["bSolution", "E", "_e"],
        "ePrimary": ["bSolution", "E", "_ePrimary"],
        "eSecondary": ["bSolution", "E", "_eSecondary"],
        "j": ["bSolution", "CCV", "_j"],
        "h": ["bSolution", "CCV", "_h"],
    }

    def startup(self):
        self.prob = self.survey.prob
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MeSigma = self.survey.prob.MeSigma
        self._MeSigmaI = self.survey.prob.MeSigmaI
        self._MfMui = self.survey.prob.MfMui
        self._MfMuiDeriv = self.survey.prob.MfMuiDeriv
        self._MeSigmaDeriv = self.survey.prob.MeSigmaDeriv
        self._MeSigmaIDeriv = self.survey.prob.MeSigmaIDeriv
        self._Me = self.survey.prob.Me
        self._aveF2CCV = self.survey.prob.mesh.aveF2CCV
        self._aveE2CCV = self.survey.prob.mesh.aveE2CCV
        self._sigma = self.survey.prob.sigma
        self._mui = self.survey.prob.mui
        self._nC = self.survey.prob.mesh.nC

    def _GLoc(self, fieldType):
        if fieldType in ["e", "eSecondary", "ePrimary"]:
            return "E"
        elif fieldType in ["b", "bSecondary", "bPrimary"]:
            return "F"
        elif (fieldType == "h") or (fieldType == "j"):
            return "CCV"
        else:
            raise Exception("Field type must be e, b, h, j")

    def _bPrimary(self, bSolution, srcList):
        """
        Primary magnetic flux density from source

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary electric field as defined by the sources
        """

        bPrimary = np.zeros([self.prob.mesh.nF, len(srcList)], dtype=complex)
        for i, src in enumerate(srcList):
            bp = src.bPrimary(self.prob)
            bPrimary[:, i] = bPrimary[:, i] + bp
        return bPrimary

    def _bSecondary(self, bSolution, srcList):
        """
        Secondary magnetic flux density is the thing we solved for

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary magnetic flux density
        """

        return bSolution

    def _bDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Partial derivative of the total magnetic flux density with respect to
        the thing we solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic flux density with
            respect to the field we solved for with a vector
        """

        return Identity() * du_dm_v

    def _bDeriv_m(self, src, v, adjoint=False):
        """
        Partial derivative of the total magnetic flux density with respect to
        the inversion model. Here, we assume that the primary does not depend
        on the model. Note that this also includes derivative contributions
        from the sources.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: SimPEG.Utils.Zero
        :return: product of the magnetic flux density derivative with respect
            to the inversion model with a vector
        """

        # assuming primary does not depend on the model
        return Zero()

    def _ePrimary(self, bSolution, srcList):
        """
        Primary electric field from source

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary electric field as defined by the sources
        """

        ePrimary = np.zeros(
            [self._edgeCurl.shape[1], bSolution.shape[1]], dtype=complex
        )
        for i, src in enumerate(srcList):
            ep = src.ePrimary(self.prob)
            ePrimary[:, i] = ePrimary[:, i] + ep
        return ePrimary

    def _eSecondary(self, bSolution, srcList):
        """
        Secondary electric field from bSolution

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary electric field
        """

        e = self._edgeCurl.T * (self._MfMui * bSolution)
        for i, src in enumerate(srcList):
            s_e = src.s_e(self.prob)
            e[:, i] = e[:, i] + -s_e

        return self._MeSigmaI * e

    def _eDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the electric field with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the field we solved for with a vector
        """

        if not adjoint:
            return self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * du_dm_v))
        return self._MfMui.T * (self._edgeCurl * (self._MeSigmaI.T * du_dm_v))

    def _eDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the electric field with respect to the inversion model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the model with a vector
        """

        bSolution = Utils.mkvc(self[src, "bSolution"])
        s_e = src.s_e(self.prob)

        w = -s_e + self._edgeCurl.T * (self._MfMui * bSolution)

        if adjoint:
            s_eDeriv = src.s_eDeriv(self.prob, self._MeSigmaI.T * v, adjoint)
            return (
                self._MeSigmaIDeriv(w).T * v
                + self._MfMuiDeriv(bSolution).T
                * (self._edgeCurl * (self._MeSigmaI.T * v))
                - s_eDeriv
                + src.ePrimaryDeriv(self.prob, v, adjoint)
            )
        s_eDeriv = src.s_eDeriv(self.prob, v, adjoint)
        return (
            self._MeSigmaIDeriv(w) * v
            + self._MeSigmaI * (self._edgeCurl.T * (self._MfMuiDeriv(bSolution) * v))
            - self._MeSigmaI * s_eDeriv
            + src.ePrimaryDeriv(self.prob, v, adjoint)
        )

    def _j(self, bSolution, srcList):
        """
        Secondary current density from bSolution

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary current density
        """

        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        j = self._edgeCurl.T * (self._MfMui * bSolution)
        for i, src in enumerate(srcList):
            s_e = src.s_e(self.prob)
            j[:, i] = j[:, i] - s_e

        return VI * (self._aveE2CCV * j)

    def _jDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Partial derivative of the current density with respect to the thing we
        solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the current density with respect
            to the field we solved for with a vector
        """
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._MfMui.T * (
                self._edgeCurl * (self._aveE2CCV.T * (VI.T * du_dm_v))
            )
        return VI * (self._aveE2CCV * (self._edgeCurl.T * (self._MfMui * du_dm_v)))
        # forgetting the source term here

    def _jDeriv_mui(self, src, v, adjoint=False):
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        MfMuiDeriv = self._MfMuiDeriv(self[src, "b"])

        if adjoint:
            return MfMuiDeriv.T * (self._edgeCurl * (self._aveE2CCV.T * (VI.T * v)))

        return VI * (self._aveE2CCV * (self._edgeCurl.T * (MfMuiDeriv * v)))

    def _jDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the current density with respect to the inversion model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the current density with respect
            to the model with a vector
        """

        return self._jDeriv_mui(src, v, adjoint)

    def _h(self, bSolution, srcList):
        """
        Magnetic field from bSolution

        :param numpy.ndarray bSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: magnetic field
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        return VI * (self._aveF2CCV * (self._MfMui * self._b(bSolution, srcList)))

    def _hDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Partial derivative of the magnetic field with respect to the thing we
        solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the field we solved for with a vector
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint:
            return self._MfMui.T * (self._aveF2CCV.T * (VI.T * du_dm_v))
        return VI * (self._aveF2CCV * (self._MfMui * du_dm_v))

    def _hDeriv_mui(self, src, v, adjoint=False):
        b = self[src, "b"]
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint:
            return self._MfMuiDeriv(b).T * (self._aveF2CCV.T * (VI * v))
        return VI * (self._aveF2CCV * (self._MfMuiDeriv(b) * v))

    def _hDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic field with respect to the inversion model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the model with a vector
        """
        return src.hPrimaryDeriv(self.prob, v, adjoint) + self._hDeriv_mui(
            src, v, adjoint
        )


class Fields3D_j(FieldsFDEM):
    """
    Fields object for Problem3D_j.

    :param discretize.BaseMesh.BaseMesh mesh: mesh
    :param SimPEG.EM.FDEM.SurveyFDEM.Survey survey: survey
    """

    knownFields = {"jSolution": "F"}
    aliasFields = {
        "j": ["jSolution", "F", "_j"],
        "jPrimary": ["jSolution", "F", "_jPrimary"],
        "jSecondary": ["jSolution", "F", "_jSecondary"],
        "h": ["jSolution", "E", "_h"],
        "hPrimary": ["jSolution", "E", "_hPrimary"],
        "hSecondary": ["jSolution", "E", "_hSecondary"],
        "e": ["jSolution", "CCV", "_e"],
        "b": ["jSolution", "CCV", "_b"],
    }

    def startup(self):
        self.prob = self.survey.prob
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MeMu = self.survey.prob.MeMu
        self._MeMuI = self.survey.prob.MeMuI
        self._MeMuIDeriv = self.survey.prob.MeMuIDeriv
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv
        self._rho = self.survey.prob.rho
        self._mu = self.survey.prob.mui
        self._aveF2CCV = self.survey.prob.mesh.aveF2CCV
        self._aveE2CCV = self.survey.prob.mesh.aveE2CCV
        self._nC = self.survey.prob.mesh.nC

    def _GLoc(self, fieldType):
        if fieldType in ["h", "hSecondary", "hPrimary"]:
            return "E"
        elif fieldType in ["j", "jSecondary", "jPrimary"]:
            return "F"
        elif (fieldType == "e") or (fieldType == "b"):
            return "CCV"
        else:
            raise Exception("Field type must be e, b, h, j")

    def _jPrimary(self, jSolution, srcList):
        """
        Primary current density from source

        :param numpy.ndarray jSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary current density as defined by the sources
        """

        jPrimary = np.zeros_like(jSolution, dtype=complex)
        for i, src in enumerate(srcList):
            jp = src.jPrimary(self.prob)
            jPrimary[:, i] = jPrimary[:, i] + jp
        return jPrimary

    def _jSecondary(self, jSolution, srcList):
        """
        Secondary current density is the thing we solved for

        :param numpy.ndarray jSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary current density
        """

        return jSolution

    def _j(self, jSolution, srcList):
        """
        Total current density is sum of primary and secondary

        :param numpy.ndarray jSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: total current density
        """

        return self._jPrimary(jSolution, srcList) + self._jSecondary(jSolution, srcList)

    def _jDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Partial derivative of the total current density with respect to the
        thing we solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the current density with respect
            to the field we solved for with a vector
        """

        return Identity() * du_dm_v

    def _jDeriv_m(self, src, v, adjoint=False):
        """
        Partial derivative of the total current density with respect to the
        inversion model. Here, we assume that the primary does not depend on
        the model. Note that this also includes derivative contributions from
        the sources.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: SimPEG.Utils.Zero
        :return: product of the current density derivative with respect to the
            inversion model with a vector
        """
        # assuming primary does not depend on the model
        return src.jPrimaryDeriv(self.prob, v, adjoint)

    def _hPrimary(self, jSolution, srcList):
        """
        Primary magnetic field from source

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary magnetic field as defined by the sources
        """

        hPrimary = np.zeros(
            [self._edgeCurl.shape[1], jSolution.shape[1]], dtype=complex
        )
        for i, src in enumerate(srcList):
            hp = src.hPrimary(self.prob)
            hPrimary[:, i] = hPrimary[:, i] + hp
        return hPrimary

    def _hSecondary(self, jSolution, srcList):
        """
        Secondary magnetic field from bSolution

        :param numpy.ndarray jSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary magnetic field
        """

        h = self._edgeCurl.T * (self._MfRho * jSolution)
        for i, src in enumerate(srcList):
            h[:, i] *= -1.0 / (1j * omega(src.freq))
            s_m = src.s_m(self.prob)
            h[:, i] = h[:, i] + 1.0 / (1j * omega(src.freq)) * (s_m)
        return self._MeMuI * h

    def _hDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the magnetic field with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the field we solved for with a vector
        """

        if adjoint:
            return (
                -1.0
                / (1j * omega(src.freq))
                * self._MfRho.T
                * (self._edgeCurl * (self._MeMuI.T * du_dm_v))
            )
        return (
            -1.0
            / (1j * omega(src.freq))
            * self._MeMuI
            * (self._edgeCurl.T * (self._MfRho * du_dm_v))
        )

    def _hDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic field with respect to the inversion model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the model with a vector
        """

        jSolution = Utils.mkvc(self[[src], "jSolution"])
        MeMuI = self._MeMuI
        MeMuIDeriv = self._MeMuIDeriv
        C = self._edgeCurl
        MfRho = self._MfRho
        MfRhoDeriv = self._MfRhoDeriv

        s_m = src.s_m(self.prob)

        def s_mDeriv(v):
            return src.s_mDeriv(self.prob, v, adjoint=adjoint)

        if not adjoint:
            hDeriv_m = (
                1.0
                / (1j * omega(src.freq))
                * (
                    -1.0
                    * (
                        MeMuI * (C.T * (MfRhoDeriv(jSolution) * v))
                        + MeMuIDeriv(C.T * (MfRho * jSolution)) * v
                    )
                    + MeMuI * s_mDeriv(v)
                    + MeMuIDeriv(s_m) * v
                )
            )

        elif adjoint:
            hDeriv_m = (
                1.0
                / (1j * omega(src.freq))
                * (
                    (
                        -1.0
                        * (
                            MfRhoDeriv(jSolution).T * (C * (MeMuI.T * v))
                            + MeMuIDeriv(C.T * (MfRho * jSolution)).T * v
                        )
                    )
                    + s_mDeriv(MeMuI.T * v)
                    + MeMuIDeriv(s_m).T * v
                )
            )

        return hDeriv_m + src.hPrimaryDeriv(self.prob, v, adjoint)

    def _e(self, jSolution, srcList):
        """
        Electric field from jSolution

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: electric field
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        return VI * (self._aveF2CCV * (self._MfRho * self._j(jSolution, srcList)))

    def _eDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the electric field with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the field we solved for with a vector
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._MfRho.T * (self._aveF2CCV.T * (VI.T * du_dm_v))
        return VI * (self._aveF2CCV * (self._MfRho * du_dm_v))

    def _eDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the electric field with respect to the inversion model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the model with a vector
        """
        jSolution = Utils.mkvc(self[src, "jSolution"])
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._MfRhoDeriv(jSolution).T * (
                self._aveF2CCV.T * (VI.T * v)
            ) + src.ePrimaryDeriv(self.prob, v, adjoint)
        return VI * (
            self._aveF2CCV * (self._MfRhoDeriv(jSolution) * v)
        ) + src.ePrimaryDeriv(self.prob, v, adjoint)

    def _b(self, jSolution, srcList):
        """
        Secondary magnetic flux density from jSolution

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary magnetic flux density
        """
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        return VI * (self._aveE2CCV * (self._MeMu * self._h(jSolution, srcList)))

    def _bDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the thing we
        solved for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic flux density with
            respect to the field we solved for with a vector
        """
        # if self.prob.mesh._meshType == 'CYL':
        #     self.
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        if adjoint:
            return (
                -1.0
                / (1j * omega(src.freq))
                * self._MfRho.T
                * (self._edgeCurl * (self._aveE2CCV.T * (VI.T * du_dm_v)))
            )

        return (
            -1.0
            / (1j * omega(src.freq))
            * VI
            * (self._aveE2CCV * (self._edgeCurl.T * (self._MfRho * du_dm_v)))
        )

    def _bDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the inversion
        model

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic flux density with
            respect to the model with a vector
        """
        jSolution = self[src, "jSolution"]
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        def s_mDeriv(v):
            return src.s_mDeriv(self.prob, v, adjoint=adjoint)

        if adjoint:
            v = self._aveE2CCV.T * (VI.T * v)
            return 1.0 / (1j * omega(src.freq)) * (
                s_mDeriv(v) - self._MfRhoDeriv(jSolution).T * (self._edgeCurl * v)
            ) + src.bPrimaryDeriv(self.prob, v, adjoint)
        return 1.0 / (1j * omega(src.freq)) * VI * (
            self._aveE2CCV
            * (s_mDeriv(v) - self._edgeCurl.T * (self._MfRhoDeriv(jSolution) * v))
        ) + src.bPrimaryDeriv(self.prob, v, adjoint)


class Fields3D_h(FieldsFDEM):
    """
    Fields object for Problem3D_h.

    :param discretize.BaseMesh.BaseMesh mesh: mesh
    :param SimPEG.EM.FDEM.SurveyFDEM.Survey survey: survey
    """

    knownFields = {"hSolution": "E"}
    aliasFields = {
        "h": ["hSolution", "E", "_h"],
        "hPrimary": ["hSolution", "E", "_hPrimary"],
        "hSecondary": ["hSolution", "E", "_hSecondary"],
        "j": ["hSolution", "F", "_j"],
        "jPrimary": ["hSolution", "F", "_jPrimary"],
        "jSecondary": ["hSolution", "F", "_jSecondary"],
        "e": ["hSolution", "CCV", "_e"],
        "b": ["hSolution", "CCV", "_b"],
    }

    def startup(self):
        self.prob = self.survey.prob
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MeMu = self.survey.prob.MeMu
        self._MeMuDeriv = self.survey.prob.MeMuDeriv
        # self._MeMuI = self.survey.prob.MeMuI
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv
        self._rho = self.survey.prob.rho
        self._mu = self.survey.prob.mui
        self._aveF2CCV = self.survey.prob.mesh.aveF2CCV
        self._aveE2CCV = self.survey.prob.mesh.aveE2CCV
        self._nC = self.survey.prob.mesh.nC

    def _GLoc(self, fieldType):
        if fieldType in ["h", "hSecondary", "hPrimary"]:
            return "E"
        elif fieldType in ["j", "jSecondary", "jPrimary"]:
            return "F"
        elif (fieldType == "e") or (fieldType == "b"):
            return "CCV"
        else:
            raise Exception("Field type must be e, b, h, j")

    def _hPrimary(self, hSolution, srcList):
        """
        Primary magnetic field from source

        :param numpy.ndarray eSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary magnetic field as defined by the sources
        """

        hPrimary = np.zeros_like(hSolution, dtype=complex)
        for i, src in enumerate(srcList):
            hp = src.hPrimary(self.prob)
            hPrimary[:, i] = hPrimary[:, i] + hp
        return hPrimary

    def _hSecondary(self, hSolution, srcList):
        """
        Secondary magnetic field is the thing we solved for

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary magnetic field
        """

        return hSolution

    def _hDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Partial derivative of the total magnetic field with respect to the
        thing we solved for.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic field with respect
            to the field we solved for with a vector
        """

        return Identity() * du_dm_v

    def _hDeriv_m(self, src, v, adjoint=False):
        """
        Partial derivative of the total magnetic field with respect to the
            inversion model. Here, we assume that the primary does not depend
            on the model. Note that this also includes derivative contributions
            from the sources.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: SimPEG.Utils.Zero
        :return: product of the magnetic field derivative with respect to the
            inversion model with a vector
        """

        return src.hPrimaryDeriv(self.prob, v, adjoint)

    def _jPrimary(self, hSolution, srcList):
        """
        Primary current density from source

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: primary current density as defined by the sources
        """

        jPrimary = np.zeros(
            [self._edgeCurl.shape[0], hSolution.shape[1]], dtype=complex
        )
        for i, src in enumerate(srcList):
            jp = src.jPrimary(self.prob)
            jPrimary[:, i] = jPrimary[:, i] + jp
        return jPrimary

    def _jSecondary(self, hSolution, srcList):
        """
        Secondary current density from hSolution

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: secondary current density
        """

        j = self._edgeCurl * hSolution
        for i, src in enumerate(srcList):
            s_e = src.s_e(self.prob)
            j[:, i] = j[:, i] + -s_e
        return j

    def _jDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the current density with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the current density with respect
            to the field we solved for with a vector
        """

        if not adjoint:
            return self._edgeCurl * du_dm_v
        elif adjoint:
            return self._edgeCurl.T * du_dm_v

    def _jDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the current density with respect to the inversion model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the current density derivative with respect to the
            inversion model with a vector
        """

        return -src.s_eDeriv(self.prob, v, adjoint) + src.jPrimaryDeriv(
            self.prob, v, adjoint
        )

    def _e(self, hSolution, srcList):
        """
        Electric field from hSolution

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: electric field
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        return VI * (self._aveF2CCV * (self._MfRho * self._j(hSolution, srcList)))

    def _eDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the electric field with respect to the thing we solved
        for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the electric field with respect
            to the field we solved for with a vector
        """
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._edgeCurl.T * (
                self._MfRho.T * (self._aveF2CCV.T * (VI.T * du_dm_v))
            )
        return VI * (self._aveF2CCV * (self._MfRho * self._edgeCurl * du_dm_v))

    def _eDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the electric field with respect to the inversion model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the electric field derivative with respect to the
            inversion model with a vector
        """
        hSolution = Utils.mkvc(self[src, "hSolution"])
        n = int(self._aveF2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        s_e = src.s_e(self.prob)

        if adjoint:
            w = self._aveF2CCV.T * (VI.T * v)
            return (
                self._MfRhoDeriv(self._edgeCurl * hSolution).T * w
                - self._MfRhoDeriv(s_e).T * w
                + src.ePrimaryDeriv(self.prob, v, adjoint)
            )
        return VI * (
            self._aveF2CCV
            * (
                self._MfRhoDeriv(self._edgeCurl * hSolution) * v
                - self._MfRhoDeriv(s_e) * v
            )
        ) + src.ePrimaryDeriv(self.prob, v, adjoint)

    def _b(self, hSolution, srcList):
        """
        Magnetic flux density from hSolution

        :param numpy.ndarray hSolution: field we solved for
        :param list srcList: list of sources
        :rtype: numpy.ndarray
        :return: magnetic flux density
        """
        h = self._h(hSolution, srcList)
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))

        return VI * (self._aveE2CCV * (self._MeMu * h))

    def _bDeriv_u(self, src, du_dm_v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the thing we
        solved for

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray du_dm_v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the derivative of the magnetic flux density with
            respect to the field we solved for with a vector
        """
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        if adjoint:
            return self._MeMu.T * (self._aveE2CCV.T * (VI.T * du_dm_v))
        return VI * (self._aveE2CCV * (self._MeMu * du_dm_v))

    def _bDeriv_mu(self, src, v, adjoint=False):
        h = self[src, "h"]
        n = int(self._aveE2CCV.shape[0] / self._nC)  # number of components
        VI = sdiag(np.kron(np.ones(n), 1.0 / self.prob.mesh.vol))
        MeMuDeriv = self._MeMuDeriv(h)

        if adjoint:
            return MeMuDeriv.T * (self._aveE2CCV.T * (VI.T * v))
        return VI * (self._aveE2CCV * (MeMuDeriv * v))
        # return VI * (self._aveE2CCV * (self._MeMu * h))

    def _bDeriv_m(self, src, v, adjoint=False):
        """
        Derivative of the magnetic flux density with respect to the inversion
        model.

        :param SimPEG.EM.FDEM.SrcFDEM.BaseFDEMSrc src: source
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of the magnetic flux density derivative with respect
            to the inversion model with a vector
        """
        return src.bPrimaryDeriv(self.prob, v, adjoint) + self._bDeriv_mu(
            src, v, adjoint
        )



from . import geophysical_systems
from .utils import (
    filter_xy,
    rotate_xy,
    find_value,
    export_grid_2_geotiff,
    geotiff_2_grid,
    export_curve_2_shapefile,
    running_mean,
    signal_processing_1d,
    tensor_2_block_model,
    block_model_2_tensor,
    treemesh_2_octree,
    octree_2_treemesh,
    object_2_dataframe,
    data_2_zarr,
    rotate_vertices,
    RectangularBlock,
    format_labels,
)



import numpy as np
from numpy.linalg import norm
from SimPEG.Utils import mkvc, sdiag, diagEst
from . import Utils
from discretize import TensorMesh, CurvilinearMesh, CylMesh
from discretize.TreeMesh import TreeMesh as Tree
import scipy.sparse as sp
import unittest
import inspect

try:
    import getpass

    name = getpass.getuser()[0].upper() + getpass.getuser()[1:]
except Exception as e:
    name = "You"
happiness = [
    "The test be workin!",
    "You get a gold star!",
    "Yay passed!",
    "Happy little convergence test!",
    "That was easy!",
    "Testing is important.",
    "You are awesome.",
    "Go Test Go!",
    "Once upon a time, a happy little test passed.",
    "And then everyone was happy.",
    "Not just a pretty face " + name,
    "You deserve a pat on the back!",
    "Well done " + name + "!",
    "Awesome, " + name + ", just awesome.",
]
sadness = [
    "No gold star for you.",
    "Try again soon.",
    "Thankfully,  persistence is a great substitute for talent.",
    "It might be easier to call this a feature...",
    "Coffee break?",
    "Boooooooo  :(",
    "Testing is important. Do it again.",
    "Did you put your clever trousers on today?",
    "Just think about a dancing dinosaur and life will get better!",
    "You had so much promise " + name + ", oh well...",
    name.upper() + " ERROR!",
    "Get on it " + name + "!",
    "You break it, you fix it.",
]


class OrderTest(unittest.TestCase):
    r"""

    OrderTest is a base class for testing convergence orders with respect to mesh
    sizes of integral/differential operators.

    Mathematical Problem:

        Given are an operator A and its discretization A[h]. For a given test function f
        and h --> 0  we compare:

        .. math::
            error(h) = \| A[h](f) - A(f) \|_{\infty}

        Note that you can provide any norm.

        Test is passed when estimated rate order of convergence is  at least within the specified tolerance of the
        estimated rate supplied by the user.

    Minimal example for a curl operator::

        class TestCURL(OrderTest):
            name = "Curl"

            def getError(self):
                # For given Mesh, generate A[h], f and A(f) and return norm of error.


                fun  = lambda x: np.cos(x)  # i (cos(y)) + j (cos(z)) + k (cos(x))
                sol = lambda x: np.sin(x)  # i (sin(z)) + j (sin(x)) + k (sin(y))


                Ex = fun(self.M.gridEx[:, 1])
                Ey = fun(self.M.gridEy[:, 2])
                Ez = fun(self.M.gridEz[:, 0])
                f = np.concatenate((Ex, Ey, Ez))

                Fx = sol(self.M.gridFx[:, 2])
                Fy = sol(self.M.gridFy[:, 0])
                Fz = sol(self.M.gridFz[:, 1])
                Af = np.concatenate((Fx, Fy, Fz))

                # Generate DIV matrix
                Ah = self.M.edgeCurl

                curlE = Ah*E
                err = np.linalg.norm((Ah*f -Af), np.inf)
                return err

            def test_order(self):
                # runs the test
                self.orderTest()

    See also: test_operatorOrder.py

    """

    name = "Order Test"
    expectedOrders = (
        2.0  # This can be a list of orders, must be the same length as meshTypes
    )
    tolerance = 0.85  # This can also be a list, must be the same length as meshTypes
    meshSizes = [4, 8, 16, 32]
    meshTypes = ["uniformTensorMesh"]
    _meshType = meshTypes[0]
    meshDimension = 3

    def setupMesh(self, nc):
        """
        For a given number of cells nc, generate a TensorMesh with uniform cells with edge length h=1/nc.
        """
        if "TensorMesh" in self._meshType:
            if "uniform" in self._meshType:
                h = [nc, nc, nc]
            elif "random" in self._meshType:
                h1 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h2 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h3 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h = [hi / np.sum(hi) for hi in [h1, h2, h3]]  # normalize
            else:
                raise Exception("Unexpected meshType")

            self.M = TensorMesh(h[: self.meshDimension])
            max_h = max([np.max(hi) for hi in self.M.h])
            return max_h

        elif "CylMesh" in self._meshType:
            if "uniform" in self._meshType:
                h = [nc, nc, nc]
            else:
                raise Exception("Unexpected meshType")

            if self.meshDimension == 2:
                self.M = CylMesh([h[0], 1, h[2]])
                max_h = max([np.max(hi) for hi in [self.M.hx, self.M.hz]])
            elif self.meshDimension == 3:
                self.M = CylMesh(h)
                max_h = max([np.max(hi) for hi in self.M.h])
            return max_h

        elif "Curv" in self._meshType:
            if "uniform" in self._meshType:
                kwrd = "rect"
            elif "rotate" in self._meshType:
                kwrd = "rotate"
            else:
                raise Exception("Unexpected meshType")
            if self.meshDimension == 1:
                raise Exception("Lom not supported for 1D")
            elif self.meshDimension == 2:
                X, Y = Utils.exampleLrmGrid([nc, nc], kwrd)
                self.M = CurvilinearMesh([X, Y])
            elif self.meshDimension == 3:
                X, Y, Z = Utils.exampleLrmGrid([nc, nc, nc], kwrd)
                self.M = CurvilinearMesh([X, Y, Z])
            return 1.0 / nc

        elif "Tree" in self._meshType:
            nc *= 2
            if "uniform" in self._meshType or "notatree" in self._meshType:
                h = [nc, nc, nc]
            elif "random" in self._meshType:
                h1 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h2 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h3 = np.random.rand(nc) * nc * 0.5 + nc * 0.5
                h = [hi / np.sum(hi) for hi in [h1, h2, h3]]  # normalize
            else:
                raise Exception("Unexpected meshType")

            levels = int(np.log(nc) / np.log(2))
            self.M = Tree(h[: self.meshDimension], levels=levels)

            def function(cell):
                if "notatree" in self._meshType:
                    return levels - 1
                r = cell.center - np.array([0.5] * len(cell.center))
                dist = np.sqrt(r.dot(r))
                if dist < 0.2:
                    return levels
                return levels - 1

            self.M.refine(function, balance=False)
            self.M.number(balance=False)
            # self.M.plotGrid(showIt=True)
            max_h = max([np.max(hi) for hi in self.M.h])
            return max_h

    def getError(self):
        """For given h, generate A[h], f and A(f) and return norm of error."""
        return 1.0

    def orderTest(self):
        """
        For number of cells specified in meshSizes setup mesh, call getError
        and prints mesh size, error, ratio between current and previous error,
        and estimated order of convergence.


        """
        assert type(self.meshTypes) == list, "meshTypes must be a list"
        if type(self.tolerance) is not list:
            self.tolerance = np.ones(len(self.meshTypes)) * self.tolerance

        # if we just provide one expected order, repeat it for each mesh type
        if type(self.expectedOrders) == float or type(self.expectedOrders) == int:
            self.expectedOrders = [self.expectedOrders for i in self.meshTypes]

        assert type(self.expectedOrders) == list, "expectedOrders must be a list"
        assert len(self.expectedOrders) == len(
            self.meshTypes
        ), "expectedOrders must have the same length as the meshTypes"

        for ii_meshType, meshType in enumerate(self.meshTypes):
            self._meshType = meshType
            self._tolerance = self.tolerance[ii_meshType]
            self._expectedOrder = self.expectedOrders[ii_meshType]

            order = []
            err_old = 0.0
            max_h_old = 0.0
            for ii, nc in enumerate(self.meshSizes):
                max_h = self.setupMesh(nc)
                err = self.getError()
                if ii == 0:
                    print("")
                    print(self._meshType + ":  " + self.name)
                    print("_____________________________________________")
                    print("   h  |    error    | e(i-1)/e(i) |  order")
                    print("~~~~~~|~~~~~~~~~~~~~|~~~~~~~~~~~~~|~~~~~~~~~~")
                    print(f"{nc:4d}  |  {err:8.2e}   |")
                else:
                    order.append(np.log(err / err_old) / np.log(max_h / max_h_old))
                    print(
                        "{:4d}  |  {:8.2e}   |   {:6.4f}    |  {:6.4f}".format(
                            nc, err, err_old / err, order[-1]
                        )
                    )
                err_old = err
                max_h_old = max_h
            print("---------------------------------------------")
            passTest = np.mean(np.array(order)) > self._tolerance * self._expectedOrder
            if passTest:
                print(happiness[np.random.randint(len(happiness))])
            else:
                print("Failed to pass test on " + self._meshType + ".")
                print(sadness[np.random.randint(len(sadness))])
            print("")
            self.assertTrue(passTest)


def Rosenbrock(x, return_g=True, return_H=True):
    """Rosenbrock function for testing GaussNewton scheme"""

    f = 100 * (x[1] - x[0] ** 2) ** 2 + (1 - x[0]) ** 2
    g = np.array(
        [2 * (200 * x[0] ** 3 - 200 * x[0] * x[1] + x[0] - 1), 200 * (x[1] - x[0] ** 2)]
    )
    H = sp.csr_matrix(
        np.array(
            [[-400 * x[1] + 1200 * x[0] ** 2 + 2, -400 * x[0]], [-400 * x[0], 200]]
        )
    )

    out = (f,)
    if return_g:
        out += (g,)
    if return_H:
        out += (H,)
    return out if len(out) > 1 else out[0]


def checkDerivative(
    fctn,
    x0,
    num=7,
    plotIt=True,
    dx=None,
    expectedOrder=2,
    tolerance=0.85,
    eps=1e-10,
    ax=None,
):
    """
        Basic derivative check

        Compares error decay of 0th and 1st order Taylor approximation at point
        x0 for a randomized search direction.

        :param callable fctn: function handle
        :param numpy.array x0: point at which to check derivative
        :param int num: number of times to reduce step length, h
        :param bool plotIt: if you would like to plot
        :param numpy.array dx: step direction
        :param int expectedOrder: The order that you expect the derivative to yield.
        :param float tolerance: The tolerance on the expected order.
        :param float eps: What is zero?
        :rtype: bool
        :return: did you pass the test?!


        .. plot::
            :include-source:

            import numpy as np
            from .. import Tests, Utils
            def simplePass(x):
                return np.sin(x), Utils.sdiag(np.cos(x))
            Tests.checkDerivative(simplePass, np.random.randn(5))
    """

    print("{!s} checkDerivative {!s}".format("=" * 20, "=" * 20))
    print("iter    h         |ft-f0|   |ft-f0-h*J0*dx|  Order\n{!s}".format("-" * 57))

    f0, J0 = fctn(x0)

    x0 = mkvc(x0)

    if dx is None:
        dx = np.random.randn(len(x0))

    h = np.logspace(-1, -num, num)
    E0 = np.ones(h.shape)
    E1 = np.ones(h.shape)

    def l2norm(x):
        # because np.norm breaks if they are scalars?
        return np.sqrt(np.real(np.vdot(x, x)))

    for i in range(num):
        # Evaluate at test point
        ft, Jt = fctn(x0 + h[i] * dx)
        # 0th order Taylor
        E0[i] = l2norm(ft - f0)
        # 1st order Taylor
        if inspect.isfunction(J0):
            E1[i] = l2norm(ft - f0 - h[i] * J0(dx))
        elif not isinstance(J0, np.ndarray):
            E1[i] = l2norm(ft - f0 - h[i] * (J0 * dx))
        else:
            # We assume it is a numpy.ndarray
            E1[i] = l2norm(ft - f0 - h[i] * J0.dot(dx))

        order0 = np.log10(E0[:-1] / E0[1:])
        order1 = np.log10(E1[:-1] / E1[1:])
        print(
            " {:d}   {:1.2e}    {:1.3e}     {:1.3e}      {:1.3f}".format(
                i, h[i], E0[i], E1[i], np.nan if i == 0 else order1[i - 1]
            )
        )

    # Ensure we are about precision
    order0 = order0[E0[1:] > eps]
    order1 = order1[E1[1:] > eps]
    belowTol = order1.size == 0 and order0.size >= 0
    # Make sure we get the correct order
    correctOrder = order1.size > 0 and np.mean(order1) > tolerance * expectedOrder

    passTest = belowTol or correctOrder

    if passTest:
        print("{!s} PASS! {!s}".format("=" * 25, "=" * 25))
        print(happiness[np.random.randint(len(happiness))] + "\n")
    else:
        print(
            "{!s}\n{!s} FAIL! {!s}\n{!s}".format("*" * 57, "<" * 25, ">" * 25, "*" * 57)
        )
        print(sadness[np.random.randint(len(sadness))] + "\n")

    if plotIt:
        import matplotlib.pyplot as plt

        ax = ax or plt.subplot(111)
        ax.loglog(h, E0, "b")
        ax.loglog(h, E1, "g--")
        ax.set_title(
            "Check Derivative - {!s}".format("PASSED :)" if passTest else "FAILED :(")
        )
        ax.set_xlabel("h")
        ax.set_ylabel("Error")
        leg = ax.legend(
            [r"$\mathcal{O}(h)$", r"$\mathcal{O}(h^2)$"],
            loc="best",
            title=r"$f(x + h\Delta x) - f(x) - h g(x) \Delta x - \mathcal{O}(h^2) = 0$",
            frameon=False,
        )
        plt.setp(leg.get_title(), fontsize=15)
        plt.show()

    return passTest


def getQuadratic(A, b, c=0):
    r"""
        Given A, b and c, this returns a quadratic, Q

        .. math::

            \mathbf{Q( x ) = 0.5 x A x + b x} + c
    """

    def Quadratic(x, return_g=True, return_H=True):
        f = 0.5 * x.dot(A.dot(x)) + b.dot(x) + c
        out = (f,)
        if return_g:
            g = A.dot(x) + b
            out += (g,)
        if return_H:
            H = A
            out += (H,)
        return out if len(out) > 1 else out[0]

    return Quadratic



import numpy as np
import SimPEG
from SimPEG.EM.Base import BaseEMSurvey
from .. import Utils
from SimPEG.EM.Static.SIP.SrcSIP import BaseSrc
from SimPEG.EM.Static.SIP.RxSIP import BaseRx
import uuid


class Survey(BaseEMSurvey):
    rxPair = BaseRx
    srcPair = BaseSrc
    times = None

    def __init__(self, srcList, **kwargs):
        self.srcList = srcList
        BaseEMSurvey.__init__(self, srcList, **kwargs)
        self.getUniqueTimes()

    def getUniqueTimes(self):
        time_rx = []
        for src in self.srcList:
            for rx in src.rxList:
                time_rx.append(rx.times)
        self.times = np.unique(np.hstack(time_rx))

    def dpred(self, m, f=None):
        """
            Predicted data.

            .. math::
                d_\\text{pred} = Pf(m)
        """
        return self.prob.forward(m, f=f)


class Data(SimPEG.Survey.Data):
    """Fancy data storage by Src and Rx"""

    def __init__(self, survey, v=None):
        self.uid = str(uuid.uuid4())
        self.survey = survey
        self._dataDict = {}
        for src in self.survey.srcList:
            self._dataDict[src] = {}
            for rx in src.rxList:
                self._dataDict[src][rx] = {}

        if v is not None:
            self.fromvec(v)

    def _ensureCorrectKey(self, key):
        if type(key) is tuple:
            if len(key) != 3:
                raise KeyError("Key must be [Src, Rx, tInd]")
            if key[0] not in self.survey.srcList:
                raise KeyError("Src Key must be a source in the survey.")
            if key[1] not in key[0].rxList:
                raise KeyError("Rx Key must be a receiver for the source.")
            return key
        elif isinstance(key, self.survey.srcPair):
            if key not in self.survey.srcList:
                raise KeyError("Key must be a source in the survey.")
            return key, None, None
        else:
            raise KeyError("Key must be [Src] or [Src,Rx] or [Src, Rx, tInd]")

    def __setitem__(self, key, value):
        src, rx, t = self._ensureCorrectKey(key)
        assert rx is not None, "set data using [Src, Rx]"
        assert isinstance(value, np.ndarray), "value must by ndarray"
        assert (
            value.size == rx.nD
        ), "value must have the same number of data as the source."
        self._dataDict[src][rx][t] = Utils.mkvc(value)

    def __getitem__(self, key):
        src, rx, t = self._ensureCorrectKey(key)
        if rx is not None:
            if rx not in self._dataDict[src]:
                raise Exception("Data for receiver has not yet been set.")
            return self._dataDict[src][rx][t]

        return np.concatenate([self[src, rx, t] for rx in src.rxList])

    def tovec(self):
        val = []
        for src in self.survey.srcList:
            for rx in src.rxList:
                for t in rx.times:
                    val.append(self[src, rx, t])
        return np.concatenate(val)

    def fromvec(self, v):
        v = Utils.mkvc(v)
        assert v.size == self.survey.nD, "v must have the correct number of data."
        indBot, indTop = 0, 0
        for src in self.survey.srcList:
            for rx in src.rxList:
                for t in rx.times:
                    indTop += rx.nRx
                    self[src, rx, t] = v[indBot:indTop]
                    indBot += rx.nRx



import numpy as np
from scipy.constants import mu_0


# TODO: make this to take a vector rather than a single frequency
def rTEfunfwd(n_layer, f, lamda, sig, chi, depth, HalfSwitch):
    """
        Compute reflection coefficients for Transverse Electric (TE) mode.
        Only one for loop for multiple layers.

        Parameters
        ----------
        n_layer : int
            The number layers
        f : complex, ndarray
            Frequency (Hz); size = (n_frequency x n_filter)
        lamda : complex, ndarray
            Frequency (Hz); size = (n_frequency x n_filter)
        sig: compelx, ndarray
            Conductivity (S/m); size = (n_layer x n_frequency x n_filter)
        chi: compelx, ndarray
            Susceptibility (SI); size = (n_layer,)
        depth: float, ndarray
            Top boundary of the layers; size = (n_ayer,)
        HalfSwitch: bool
            Switch for halfspace

        Returns
        -------
        rTE: compex, ndarray
            Reflection coefficients;
            size = (n_frequency x n_lamba)
    """

    n_frequency, n_filter = lamda.shape

    Mtemp00 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp10 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp01 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp11 = np.zeros((n_frequency, n_filter), dtype=complex)

    M1sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

    thick = -np.diff(depth)
    w = 2 * np.pi * f

    rTE = np.zeros((n_frequency, n_filter), dtype=complex)
    utemp0 = np.zeros((n_frequency, n_filter), dtype=complex)
    utemp1 = np.zeros((n_frequency, n_filter), dtype=complex)
    const = np.zeros((n_frequency, n_filter), dtype=complex)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0, :, :])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    Mtemp00 = 0.5 * (1 + const)
    Mtemp10 = 0.5 * (1 - const)
    Mtemp01 = 0.5 * (1 - const)
    Mtemp11 = 0.5 * (1 + const)

    # may be store these and reuse for sensitivity?
    M00 = []
    M10 = []
    M01 = []
    M11 = []

    M0sum00 = Mtemp00
    M0sum10 = Mtemp10
    M0sum01 = Mtemp01
    M0sum11 = Mtemp11

    if HalfSwitch:

        M1sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
        M1sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
        M1sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
        M1sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

        M1sum00 = M0sum00
        M1sum10 = M0sum10
        M1sum01 = M0sum01
        M1sum11 = M0sum11

    else:

        for j in range(n_layer - 1):
            utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j]) * sig[j, :, :])
            utemp1 = np.sqrt(
                lamda ** 2 + 1j * w * mu_0 * (1 + chi[j + 1]) * sig[j + 1, :, :]
            )
            const = mu_0 * (1 + chi[j]) * utemp1 / (mu_0 * (1 + chi[j + 1]) * utemp0)

            h0 = thick[j]

            Mtemp00 = 0.5 * (1.0 + const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp10 = 0.5 * (1.0 - const)
            Mtemp01 = 0.5 * (1.0 - const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp11 = 0.5 * (1.0 + const)

            M1sum00 = M0sum00 * Mtemp00 + M0sum01 * Mtemp10
            M1sum10 = M0sum10 * Mtemp00 + M0sum11 * Mtemp10
            M1sum01 = M0sum00 * Mtemp01 + M0sum01 * Mtemp11
            M1sum11 = M0sum10 * Mtemp01 + M0sum11 * Mtemp11

            M0sum00 = M1sum00
            M0sum10 = M1sum10
            M0sum01 = M1sum01
            M0sum11 = M1sum11

    rTE = M1sum01 / M1sum11

    return rTE


def matmul(a00, a10, a01, a11, b00, b10, b01, b11):
    """
        Compute 2x2 matrix mutiplication in vector way
        C = A*B
        C = [a00   a01] * [b00   b01] = [c00   c01]
        [a10   a11]   [b10   b11]   [c10   c11]
    """

    c00 = a00 * b00 + a01 * b10
    c10 = a10 * b00 + a11 * b10
    c01 = a00 * b01 + a01 * b11
    c11 = a10 * b01 + a11 * b11

    return c00, c10, c01, c11


# TODO: make this to take a vector rather than a single frequency
def rTEfunjac(n_layer, f, lamda, sig, chi, depth, HalfSwitch):
    """
        Compute sensitivity of reflection coefficients for
        Transverse Electric (TE) mode with regard to conductivity

        Parameters
        ----------
        n_layer : int
            The number layers
        f : complex, ndarray
            Frequency (Hz); size = (n_frequency x n_finlter)
        lamda : complex, ndarray
            Frequency (Hz); size = (n_frequency x n_finlter)
        sig: compelx, ndarray
            Conductivity (S/m); size = (n_layer x 1)
        chi: compelx, ndarray
            Susceptibility (SI); size = (n_layer x 1)
        depth: float, ndarray
            Top boundary of the layers
        HalfSwitch: bool
            Switch for halfspace

        Returns
        -------
        rTE: compex, ndarray
            Derivative of reflection coefficients;
            size = (n_frequency x n_layer x n_finlter)
    """
    # Initializing arrays
    n_frequency, n_filter = lamda.shape

    Mtemp00 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp10 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp01 = np.zeros((n_frequency, n_filter), dtype=complex)
    Mtemp11 = np.zeros((n_frequency, n_filter), dtype=complex)

    M1sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
    M1sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

    M0sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
    M0sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
    M0sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
    M0sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

    dMtemp00 = np.zeros((n_frequency, n_filter), dtype=complex)
    dMtemp10 = np.zeros((n_frequency, n_filter), dtype=complex)
    dMtemp01 = np.zeros((n_frequency, n_filter), dtype=complex)
    dMtemp11 = np.zeros((n_frequency, n_filter), dtype=complex)

    dj0temp00 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj0temp10 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj0temp01 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj0temp11 = np.zeros((n_frequency, n_filter), dtype=complex)

    dj1temp00 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj1temp10 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj1temp01 = np.zeros((n_frequency, n_filter), dtype=complex)
    dj1temp11 = np.zeros((n_frequency, n_filter), dtype=complex)

    thick = -np.diff(depth)
    w = 2 * np.pi * f

    rTE = np.zeros((n_frequency, n_filter), dtype=complex)
    drTE = np.zeros((n_layer, n_frequency, n_filter), dtype=complex)
    utemp0 = np.zeros((n_frequency, n_filter), dtype=complex)
    utemp1 = np.zeros((n_frequency, n_filter), dtype=complex)
    const = np.zeros((n_frequency, n_filter), dtype=complex)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0, :, :])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    # Compute M1
    Mtemp00 = 0.5 * (1 + const)
    Mtemp10 = 0.5 * (1 - const)
    Mtemp01 = 0.5 * (1 - const)
    Mtemp11 = 0.5 * (1 + const)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0, :, :])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    # Compute dM1du1
    dj0Mtemp00 = 0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp10 = -0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp01 = -0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp11 = 0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))

    # TODO: for computing Jacobian
    M00 = []
    M10 = []
    M01 = []
    M11 = []

    dJ00 = []
    dJ10 = []
    dJ01 = []
    dJ11 = []

    M00.append(Mtemp00)
    M01.append(Mtemp01)
    M10.append(Mtemp10)
    M11.append(Mtemp11)

    M0sum00 = Mtemp00.copy()
    M0sum10 = Mtemp10.copy()
    M0sum01 = Mtemp01.copy()
    M0sum11 = Mtemp11.copy()

    if HalfSwitch or n_layer == 1:

        M1sum00 = M0sum00.copy()
        M1sum10 = M0sum10.copy()
        M1sum01 = M0sum01.copy()
        M1sum11 = M0sum11.copy()

    else:

        for j in range(n_layer - 1):

            dJ_10Mtemp00 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ_10Mtemp10 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ_10Mtemp01 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ_10Mtemp11 = np.zeros((n_frequency, n_filter), dtype=complex)

            dJ01Mtemp00 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ01Mtemp10 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ01Mtemp01 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ01Mtemp11 = np.zeros((n_frequency, n_filter), dtype=complex)

            utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j]) * sig[j, :, :])
            utemp1 = np.sqrt(
                lamda ** 2 + 1j * w * mu_0 * (1 + chi[j + 1]) * sig[j + 1, :, :]
            )
            const = mu_0 * (1 + chi[j]) * utemp1 / (mu_0 * (1 + chi[j + 1]) * utemp0)

            h0 = thick[j]

            Mtemp00 = 0.5 * (1.0 + const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp10 = 0.5 * (1.0 - const)
            Mtemp01 = 0.5 * (1.0 - const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp11 = 0.5 * (1.0 + const)

            M1sum00, M1sum10, M1sum01, M1sum11 = matmul(
                M0sum00, M0sum10, M0sum01, M0sum11, Mtemp00, Mtemp10, Mtemp01, Mtemp11
            )

            M0sum00 = M1sum00
            M0sum10 = M1sum10
            M0sum01 = M1sum01
            M0sum11 = M1sum11

            # TODO: for Computing Jacobian

            dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[j]) / utemp0

            if j == 0:

                const1a = (
                    mu_0
                    * (1 + chi[j])
                    * utemp1
                    / (mu_0 * (1 + chi[j + 1]) * utemp0 ** 2)
                )
                const1b = const1a * utemp0

                dj1Mtemp00 = -0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 + const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp10 = 0.5 * const1a
                dj1Mtemp01 = 0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 - const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp11 = -0.5 * const1a

                # Compute dM1dm1*M2
                dJ_10Mtemp00, dJ_10Mtemp10, dJ_10Mtemp01, dJ_10Mtemp11 = matmul(
                    dj0Mtemp00,
                    dj0Mtemp10,
                    dj0Mtemp01,
                    dj0Mtemp11,
                    Mtemp00,
                    Mtemp10,
                    Mtemp01,
                    Mtemp11,
                )

                # Compute M1*dM2dm1
                dJ01Mtemp00, dJ01Mtemp10, dJ01Mtemp01, dJ01Mtemp11 = matmul(
                    M00[j],
                    M10[j],
                    M01[j],
                    M11[j],
                    dj1Mtemp00,
                    dj1Mtemp10,
                    dj1Mtemp01,
                    dj1Mtemp11,
                )

                dJ00.append(dudsig * (dJ_10Mtemp00 + dJ01Mtemp00))
                dJ10.append(dudsig * (dJ_10Mtemp10 + dJ01Mtemp10))
                dJ01.append(dudsig * (dJ_10Mtemp01 + dJ01Mtemp01))
                dJ11.append(dudsig * (dJ_10Mtemp11 + dJ01Mtemp11))

            else:

                h_1 = thick[j - 1]
                utemp_1 = np.sqrt(
                    lamda ** 2 + 1j * w * mu_0 * (1 + chi[j - 1]) * sig[j - 1]
                )
                const0 = mu_0 * (1 + chi[j - 1]) / (mu_0 * (1 + chi[j]) * utemp_1)

                dj0Mtemp00 = 0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
                dj0Mtemp10 = -0.5 * (const0)
                dj0Mtemp01 = -0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
                dj0Mtemp11 = 0.5 * (const0)

                const1a = (
                    mu_0
                    * (1 + chi[j])
                    * utemp1
                    / (mu_0 * (1 + chi[j + 1]) * utemp0 ** 2)
                )
                const1b = const1a * utemp0

                dj1Mtemp00 = -0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 + const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp10 = 0.5 * const1a
                dj1Mtemp01 = 0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 - const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp11 = -0.5 * const1a

                # Compute dMjdmj*Mj+1
                dJ_10Mtemp00, dJ_10Mtemp10, dJ_10Mtemp01, dJ_10Mtemp11 = matmul(
                    dj0Mtemp00,
                    dj0Mtemp10,
                    dj0Mtemp01,
                    dj0Mtemp11,
                    Mtemp00,
                    Mtemp10,
                    Mtemp01,
                    Mtemp11,
                )

                # Compute Mj*dMj+1dmj
                dJ01Mtemp00, dJ01Mtemp10, dJ01Mtemp01, dJ01Mtemp11 = matmul(
                    M00[j],
                    M10[j],
                    M01[j],
                    M11[j],
                    dj1Mtemp00,
                    dj1Mtemp10,
                    dj1Mtemp01,
                    dj1Mtemp11,
                )

                dJ00.append(dudsig * (dJ_10Mtemp00 + dJ01Mtemp00))
                dJ10.append(dudsig * (dJ_10Mtemp10 + dJ01Mtemp10))
                dJ01.append(dudsig * (dJ_10Mtemp01 + dJ01Mtemp01))
                dJ11.append(dudsig * (dJ_10Mtemp11 + dJ01Mtemp11))

            M00.append(Mtemp00)
            M01.append(Mtemp01)
            M10.append(Mtemp10)
            M11.append(Mtemp11)

    # rTE = M1sum01/M1sum11

    if HalfSwitch or n_layer == 1:

        utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0])
        dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[0]) / utemp0

        dJ1sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
        dJ1sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
        dJ1sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
        dJ1sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

        dJ1sum00 = dudsig * dj0Mtemp00
        dJ1sum10 = dudsig * dj0Mtemp10
        dJ1sum01 = dudsig * dj0Mtemp01
        dJ1sum11 = dudsig * dj0Mtemp11

        drTE = dJ1sum01 / M1sum11 - M1sum01 / (M1sum11 ** 2) * dJ1sum11

    else:

        # j = n_layer
        utemp0 = np.sqrt(
            lamda ** 2 + 1j * w * mu_0 * (1 + chi[n_layer - 1]) * sig[n_layer - 1, :, :]
        )
        dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[n_layer - 1]) / utemp0

        h_1 = thick[n_layer - 2]

        utemp_1 = np.sqrt(
            lamda ** 2 + 1j * w * mu_0 * (1 + chi[n_layer - 2]) * sig[n_layer - 2, :, :]
        )
        const0 = (
            mu_0 * (1 + chi[n_layer - 2]) / (mu_0 * (1 + chi[n_layer - 1]) * utemp_1)
        )

        dj0Mtemp00 = 0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
        dj0Mtemp10 = -0.5 * (const0)
        dj0Mtemp01 = -0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
        dj0Mtemp11 = 0.5 * (const0)

        dJ_10Mtemp00 = dj0Mtemp00
        dJ_10Mtemp10 = dj0Mtemp10
        dJ_10Mtemp01 = dj0Mtemp01
        dJ_10Mtemp11 = dj0Mtemp11

        dJ00.append(dudsig * dJ_10Mtemp00)
        dJ10.append(dudsig * dJ_10Mtemp10)
        dJ01.append(dudsig * dJ_10Mtemp01)
        dJ11.append(dudsig * dJ_10Mtemp11)

        for i in range(n_layer):

            dJ0sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ0sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ0sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ0sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

            dJ1sum00 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ1sum10 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ1sum01 = np.zeros((n_frequency, n_filter), dtype=complex)
            dJ1sum11 = np.zeros((n_frequency, n_filter), dtype=complex)

            if i == 0:

                for j in range(n_layer - 2):

                    if j == 0:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    else:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            elif (i > 0) & (i < n_layer - 1):

                dJ0sum00 = M00[0]
                dJ0sum10 = M10[0]
                dJ0sum01 = M01[0]
                dJ0sum11 = M11[0]

                for j in range(n_layer - 2):

                    if j == i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                        )

                    elif j < i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 1],
                            M10[j + 1],
                            M01[j + 1],
                            M11[j + 1],
                        )

                    elif j > i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            elif i == n_layer - 1:

                dJ0sum00 = M00[0]
                dJ0sum10 = M10[0]
                dJ0sum01 = M01[0]
                dJ0sum11 = M11[0]

                for j in range(n_layer - 1):

                    if j < n_layer - 2:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 1],
                            M10[j + 1],
                            M01[j + 1],
                            M11[j + 1],
                        )

                    elif j == n_layer - 2:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            drTE[i, :] = dJ1sum01 / M1sum11 - M1sum01 / (M1sum11 ** 2) * dJ1sum11

    return drTE
    # Still worthwhile to output both?
    # return rTE, drTE



from .StraightRayProblem import StraightRayProblem as Problem
from .StraightRayProblem import lengthInCell
from .StraightRaySurvey import StraightRaySurvey as Survey
from ...Survey import BaseSrc as Src
from ...Survey import BaseRx as Rx



import SimPEG
from SimPEG.Utils import Identity, Zero
import numpy as np
from scipy.constants import epsilon_0


class FieldsDC(SimPEG.Problem.Fields):
    knownFields = {}
    dtype = float

    def _phiDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_phiDeriv_u", None) is None
            or getattr(self, "_phiDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting phiDerivs from {!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (
                self._phiDeriv_u(src, v, adjoint=adjoint),
                self._phiDeriv_m(src, v, adjoint=adjoint),
            )

        return np.array(
            self._phiDeriv_u(src, du_dm_v, adjoint) + self._phiDeriv_m(src, v, adjoint),
            dtype=float,
        )

    def _eDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_eDeriv_u", None) is None
            or getattr(self, "_eDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting eDerivs from {!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._eDeriv_u(src, v, adjoint), self._eDeriv_m(src, v, adjoint))
        return np.array(
            self._eDeriv_u(src, du_dm_v, adjoint) + self._eDeriv_m(src, v, adjoint),
            dtype=float,
        )

    def _jDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_jDeriv_u", None) is None
            or getattr(self, "_jDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting jDerivs from {!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._jDeriv_u(src, v, adjoint), self._jDeriv_m(src, v, adjoint))
        return np.array(
            self._jDeriv_u(src, du_dm_v, adjoint) + self._jDeriv_m(src, v, adjoint),
            dtype=float,
        )


class Fields_CC(FieldsDC):
    knownFields = {"phiSolution": "CC"}
    aliasFields = {
        "phi": ["phiSolution", "CC", "_phi"],
        "j": ["phiSolution", "F", "_j"],
        "e": ["phiSolution", "F", "_e"],
        "charge": ["phiSolution", "CC", "_charge"],
    }
    # primary - secondary
    # CC variables

    def __init__(self, mesh, survey, **kwargs):
        FieldsDC.__init__(self, mesh, survey, **kwargs)
        if self.mesh._meshType == "TREE":
            if self.prob.bc_type == "Neumann":
                raise NotImplementedError()
            elif self.prob.bc_type == "Dirchlet":
                self.cellGrad = -mesh.faceDiv.T
        else:
            mesh.setCellGradBC("neumann")
            self.cellGrad = mesh.cellGrad

    def startup(self):
        self.prob = self.survey.prob

    def _GLoc(self, fieldType):
        if fieldType == "phi":
            return "CC"
        elif fieldType == "e" or fieldType == "j":
            return "F"
        else:
            raise Exception("Field type must be phi, e, j")

    def _phi(self, phiSolution, srcList):
        return phiSolution

    def _phiDeriv_u(self, src, v, adjoint=False):
        return Identity() * v

    def _phiDeriv_m(self, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        """
            .. math::
                \\mathbf{j} = \\mathbf{M}^{f \\ -1}_{\rho} \\mathbf{G} \\phi
        """
        return self.prob.MfRhoI * self.prob.Grad * phiSolution

    def _e(self, phiSolution, srcList):
        """
            In HJ formulation e is not well-defined!!
            .. math::
                \vec{e} = -\nabla \\phi
        """
        return -self.cellGrad * phiSolution

    def _charge(self, phiSolution, srcList):
        """
            .. math::
                \\int \nabla \\codt \vec{e} =  \\int \frac{\rho_v }{\\epsillon_0}
        """
        return (
            epsilon_0
            * self.prob.Vol
            * (self.mesh.faceDiv * self._e(phiSolution, srcList))
        )


class Fields_N(FieldsDC):
    knownFields = {"phiSolution": "N"}
    aliasFields = {
        "phi": ["phiSolution", "N", "_phi"],
        "j": ["phiSolution", "E", "_j"],
        "e": ["phiSolution", "E", "_e"],
        "charge": ["phiSolution", "N", "_charge"],
    }
    # primary - secondary
    # N variables

    def __init__(self, mesh, survey, **kwargs):
        FieldsDC.__init__(self, mesh, survey, **kwargs)

    def startup(self):
        self.prob = self.survey.prob

    def _GLoc(self, fieldType):
        if fieldType == "phi":
            return "N"
        elif fieldType == "e" or fieldType == "j":
            return "E"
        else:
            raise Exception("Field type must be phi, e, j")

    def _phi(self, phiSolution, srcList):
        return phiSolution

    def _phiDeriv_u(self, src, v, adjoint=False):
        return Identity() * v

    def _phiDeriv_m(self, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        r"""
            In EB formulation j is not well-defined!!
            .. math::
                \mathbf{j} = - \mathbf{M}^{e}_{\sigma} \mathbf{G} \phi
        """
        return self.prob.MeI * self.prob.MeSigma * self._e(phiSolution, srcList)

    def _e(self, phiSolution, srcList):
        """
            In HJ formulation e is not well-defined!!
            .. math::
                \vec{e} = -\nabla \\phi
        """
        return -self.mesh.nodalGrad * phiSolution

    def _charge(self, phiSolution, srcList):
        """
            .. math::
                \\int \nabla \\codt \vec{e} =  \\int \frac{\rho_v }{\\epsillon_0}
        """
        return -epsilon_0 * (
            self.mesh.nodalGrad.T
            * self.mesh.getEdgeInnerProduct()
            * self._e(phiSolution, srcList)
        )



import numpy as np
from scipy.constants import mu_0


# TODO: make this to take a vector rather than a single frequency
def rTEfunfwd(nlay, f, lamda, sig, chi, depth, HalfSwitch):
    """
        Compute reflection coefficients for Transverse Electric (TE) mode.
        Only one for loop for multiple layers. Do not use for loop for lambda,
        which has 801 times of loops (actually, this makes the code really slow).
    """
    Mtemp00 = np.zeros(lamda.size, dtype=complex)
    Mtemp10 = np.zeros(lamda.size, dtype=complex)
    Mtemp01 = np.zeros(lamda.size, dtype=complex)
    Mtemp11 = np.zeros(lamda.size, dtype=complex)

    M1sum00 = np.zeros(lamda.size, dtype=complex)
    M1sum10 = np.zeros(lamda.size, dtype=complex)
    M1sum01 = np.zeros(lamda.size, dtype=complex)
    M1sum11 = np.zeros(lamda.size, dtype=complex)

    thick = -np.diff(depth)
    w = 2 * np.pi * f

    rTE = np.zeros(lamda.size, dtype=complex)
    utemp0 = np.zeros(lamda.size, dtype=complex)
    utemp1 = np.zeros(lamda.size, dtype=complex)
    const = np.zeros(lamda.size, dtype=complex)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    Mtemp00 = 0.5 * (1 + const)
    Mtemp10 = 0.5 * (1 - const)
    Mtemp01 = 0.5 * (1 - const)
    Mtemp11 = 0.5 * (1 + const)

    M00 = []
    M10 = []
    M01 = []
    M11 = []

    M0sum00 = Mtemp00
    M0sum10 = Mtemp10
    M0sum01 = Mtemp01
    M0sum11 = Mtemp11

    if HalfSwitch == True:

        M1sum00 = np.zeros(lamda.size, dtype=complex)
        M1sum10 = np.zeros(lamda.size, dtype=complex)
        M1sum01 = np.zeros(lamda.size, dtype=complex)
        M1sum11 = np.zeros(lamda.size, dtype=complex)

        M1sum00 = M0sum00
        M1sum10 = M0sum10
        M1sum01 = M0sum01
        M1sum11 = M0sum11

    else:

        for j in range(nlay - 1):
            utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j]) * sig[j])
            utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j + 1]) * sig[j + 1])
            const = mu_0 * (1 + chi[j]) * utemp1 / (mu_0 * (1 + chi[j + 1]) * utemp0)

            h0 = thick[j]

            Mtemp00 = 0.5 * (1.0 + const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp10 = 0.5 * (1.0 - const)
            Mtemp01 = 0.5 * (1.0 - const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp11 = 0.5 * (1.0 + const)

            M1sum00 = M0sum00 * Mtemp00 + M0sum01 * Mtemp10
            M1sum10 = M0sum10 * Mtemp00 + M0sum11 * Mtemp10
            M1sum01 = M0sum00 * Mtemp01 + M0sum01 * Mtemp11
            M1sum11 = M0sum10 * Mtemp01 + M0sum11 * Mtemp11

            M0sum00 = M1sum00
            M0sum10 = M1sum10
            M0sum01 = M1sum01
            M0sum11 = M1sum11

    rTE = M1sum01 / M1sum11

    return rTE


def matmul(a00, a10, a01, a11, b00, b10, b01, b11):
    """
        Compute 2x2 matrix mutiplication in vector way
        C = A*B
        C = [a00   a01] * [b00   b01] = [c00   c01]
        [a10   a11]   [b10   b11]   [c10   c11]
    """

    c00 = a00 * b00 + a01 * b10
    c10 = a10 * b00 + a11 * b10
    c01 = a00 * b01 + a01 * b11
    c11 = a10 * b01 + a11 * b11

    return c00, c10, c01, c11


# TODO: make this to take a vector rather than a single frequency
def rTEfunjac(nlay, f, lamda, sig, chi, depth, HalfSwitch):
    """
        Compute reflection coefficients for Transverse Electric (TE) mode.
        Only one for loop for multiple layers. Do not use for loop for lambda,
        which has 801 times of loops (actually, this makes the code really slow).
    """
    # Initializing arrays
    Mtemp00 = np.zeros(lamda.size, dtype=complex)
    Mtemp10 = np.zeros(lamda.size, dtype=complex)
    Mtemp01 = np.zeros(lamda.size, dtype=complex)
    Mtemp11 = np.zeros(lamda.size, dtype=complex)

    M1sum00 = np.zeros(lamda.size, dtype=complex)
    M1sum10 = np.zeros(lamda.size, dtype=complex)
    M1sum01 = np.zeros(lamda.size, dtype=complex)
    M1sum11 = np.zeros(lamda.size, dtype=complex)

    M0sum00 = np.zeros(lamda.size, dtype=complex)
    M0sum10 = np.zeros(lamda.size, dtype=complex)
    M0sum01 = np.zeros(lamda.size, dtype=complex)
    M0sum11 = np.zeros(lamda.size, dtype=complex)

    dMtemp00 = np.zeros(lamda.size, dtype=complex)
    dMtemp10 = np.zeros(lamda.size, dtype=complex)
    dMtemp01 = np.zeros(lamda.size, dtype=complex)
    dMtemp11 = np.zeros(lamda.size, dtype=complex)

    dj0temp00 = np.zeros(lamda.size, dtype=complex)
    dj0temp10 = np.zeros(lamda.size, dtype=complex)
    dj0temp01 = np.zeros(lamda.size, dtype=complex)
    dj0temp11 = np.zeros(lamda.size, dtype=complex)

    dj1temp00 = np.zeros(lamda.size, dtype=complex)
    dj1temp10 = np.zeros(lamda.size, dtype=complex)
    dj1temp01 = np.zeros(lamda.size, dtype=complex)
    dj1temp11 = np.zeros(lamda.size, dtype=complex)

    thick = -np.diff(depth)
    w = 2 * np.pi * f

    rTE = np.zeros(lamda.size, dtype=complex)
    drTE = np.zeros((nlay, lamda.size), dtype=complex)
    utemp0 = np.zeros(lamda.size, dtype=complex)
    utemp1 = np.zeros(lamda.size, dtype=complex)
    const = np.zeros(lamda.size, dtype=complex)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    # Compute M1
    Mtemp00 = 0.5 * (1 + const)
    Mtemp10 = 0.5 * (1 - const)
    Mtemp01 = 0.5 * (1 - const)
    Mtemp11 = 0.5 * (1 + const)

    utemp0 = lamda
    utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0])
    const = mu_0 * utemp1 / (mu_0 * (1 + chi[0]) * utemp0)

    # Compute dM1du1
    dj0Mtemp00 = 0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp10 = -0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp01 = -0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))
    dj0Mtemp11 = 0.5 * (mu_0 / (mu_0 * (1 + chi[0]) * utemp0))

    # TODO: for computing Jacobian
    M00 = []
    M10 = []
    M01 = []
    M11 = []

    dJ00 = []
    dJ10 = []
    dJ01 = []
    dJ11 = []

    M00.append(Mtemp00)
    M01.append(Mtemp01)
    M10.append(Mtemp10)
    M11.append(Mtemp11)

    M0sum00 = Mtemp00.copy()
    M0sum10 = Mtemp10.copy()
    M0sum01 = Mtemp01.copy()
    M0sum11 = Mtemp11.copy()

    if HalfSwitch == True:

        M1sum00 = np.zeros(lamda.size, dtype=complex)
        M1sum10 = np.zeros(lamda.size, dtype=complex)
        M1sum01 = np.zeros(lamda.size, dtype=complex)
        M1sum11 = np.zeros(lamda.size, dtype=complex)

        M1sum00 = M0sum00.copy()
        M1sum10 = M0sum10.copy()
        M1sum01 = M0sum01.copy()
        M1sum11 = M0sum11.copy()

    else:

        for j in range(nlay - 1):

            dJ_10Mtemp00 = np.zeros(lamda.size, dtype=complex)
            dJ_10Mtemp10 = np.zeros(lamda.size, dtype=complex)
            dJ_10Mtemp01 = np.zeros(lamda.size, dtype=complex)
            dJ_10Mtemp11 = np.zeros(lamda.size, dtype=complex)

            dJ01Mtemp00 = np.zeros(lamda.size, dtype=complex)
            dJ01Mtemp10 = np.zeros(lamda.size, dtype=complex)
            dJ01Mtemp01 = np.zeros(lamda.size, dtype=complex)
            dJ01Mtemp11 = np.zeros(lamda.size, dtype=complex)

            utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j]) * sig[j])
            utemp1 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[j + 1]) * sig[j + 1])
            const = mu_0 * (1 + chi[j]) * utemp1 / (mu_0 * (1 + chi[j + 1]) * utemp0)

            h0 = thick[j]

            Mtemp00 = 0.5 * (1.0 + const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp10 = 0.5 * (1.0 - const)
            Mtemp01 = 0.5 * (1.0 - const) * np.exp(-2.0 * utemp0 * h0)
            Mtemp11 = 0.5 * (1.0 + const)

            M1sum00, M1sum10, M1sum01, M1sum11 = matmul(
                M0sum00, M0sum10, M0sum01, M0sum11, Mtemp00, Mtemp10, Mtemp01, Mtemp11
            )

            M0sum00 = M1sum00
            M0sum10 = M1sum10
            M0sum01 = M1sum01
            M0sum11 = M1sum11

            # TODO: for Computing Jacobian

            dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[j]) / utemp0

            if j == 0:

                const1a = (
                    mu_0
                    * (1 + chi[j])
                    * utemp1
                    / (mu_0 * (1 + chi[j + 1]) * utemp0 ** 2)
                )
                const1b = const1a * utemp0

                dj1Mtemp00 = -0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 + const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp10 = 0.5 * const1a
                dj1Mtemp01 = 0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 - const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp11 = -0.5 * const1a

                # Compute dM1dm1*M2
                dJ_10Mtemp00, dJ_10Mtemp10, dJ_10Mtemp01, dJ_10Mtemp11 = matmul(
                    dj0Mtemp00,
                    dj0Mtemp10,
                    dj0Mtemp01,
                    dj0Mtemp11,
                    Mtemp00,
                    Mtemp10,
                    Mtemp01,
                    Mtemp11,
                )

                # Compute M1*dM2dm1
                dJ01Mtemp00, dJ01Mtemp10, dJ01Mtemp01, dJ01Mtemp11 = matmul(
                    M00[j],
                    M10[j],
                    M01[j],
                    M11[j],
                    dj1Mtemp00,
                    dj1Mtemp10,
                    dj1Mtemp01,
                    dj1Mtemp11,
                )

                dJ00.append(dudsig * (dJ_10Mtemp00 + dJ01Mtemp00))
                dJ10.append(dudsig * (dJ_10Mtemp10 + dJ01Mtemp10))
                dJ01.append(dudsig * (dJ_10Mtemp01 + dJ01Mtemp01))
                dJ11.append(dudsig * (dJ_10Mtemp11 + dJ01Mtemp11))

            else:

                h_1 = thick[j - 1]
                utemp_1 = np.sqrt(
                    lamda ** 2 + 1j * w * mu_0 * (1 + chi[j - 1]) * sig[j - 1]
                )
                const0 = mu_0 * (1 + chi[j - 1]) / (mu_0 * (1 + chi[j]) * utemp_1)

                dj0Mtemp00 = 0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
                dj0Mtemp10 = -0.5 * (const0)
                dj0Mtemp01 = -0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
                dj0Mtemp11 = 0.5 * (const0)

                const1a = (
                    mu_0
                    * (1 + chi[j])
                    * utemp1
                    / (mu_0 * (1 + chi[j + 1]) * utemp0 ** 2)
                )
                const1b = const1a * utemp0

                dj1Mtemp00 = -0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 + const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp10 = 0.5 * const1a
                dj1Mtemp01 = 0.5 * const1a * np.exp(-2.0 * utemp0 * h0) - h0 * (
                    1 - const1b
                ) * np.exp(-2.0 * utemp0 * h0)
                dj1Mtemp11 = -0.5 * const1a

                # Compute dMjdmj*Mj+1
                dJ_10Mtemp00, dJ_10Mtemp10, dJ_10Mtemp01, dJ_10Mtemp11 = matmul(
                    dj0Mtemp00,
                    dj0Mtemp10,
                    dj0Mtemp01,
                    dj0Mtemp11,
                    Mtemp00,
                    Mtemp10,
                    Mtemp01,
                    Mtemp11,
                )

                # Compute Mj*dMj+1dmj
                dJ01Mtemp00, dJ01Mtemp10, dJ01Mtemp01, dJ01Mtemp11 = matmul(
                    M00[j],
                    M10[j],
                    M01[j],
                    M11[j],
                    dj1Mtemp00,
                    dj1Mtemp10,
                    dj1Mtemp01,
                    dj1Mtemp11,
                )

                dJ00.append(dudsig * (dJ_10Mtemp00 + dJ01Mtemp00))
                dJ10.append(dudsig * (dJ_10Mtemp10 + dJ01Mtemp10))
                dJ01.append(dudsig * (dJ_10Mtemp01 + dJ01Mtemp01))
                dJ11.append(dudsig * (dJ_10Mtemp11 + dJ01Mtemp11))

            M00.append(Mtemp00)
            M01.append(Mtemp01)
            M10.append(Mtemp10)
            M11.append(Mtemp11)

    # rTE = M1sum01/M1sum11

    if HalfSwitch == True:

        utemp0 = np.sqrt(lamda ** 2 + 1j * w * mu_0 * (1 + chi[0]) * sig[0])
        dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[0]) / utemp0

        dJ1sum00 = np.zeros(lamda.size, dtype=complex)
        dJ1sum10 = np.zeros(lamda.size, dtype=complex)
        dJ1sum01 = np.zeros(lamda.size, dtype=complex)
        dJ1sum11 = np.zeros(lamda.size, dtype=complex)

        dJ1sum00 = dudsig * dj0Mtemp00
        dJ1sum10 = dudsig * dj0Mtemp10
        dJ1sum01 = dudsig * dj0Mtemp01
        dJ1sum11 = dudsig * dj0Mtemp11

        drTE = dJ1sum01 / M1sum11 - M1sum01 / (M1sum11 ** 2) * dJ1sum11

    else:

        # j = nlay
        utemp0 = np.sqrt(
            lamda ** 2 + 1j * w * mu_0 * (1 + chi[nlay - 1]) * sig[nlay - 1]
        )
        dudsig = 0.5 * 1j * w * mu_0 * (1 + chi[j]) / utemp0

        h_1 = thick[nlay - 2]
        utemp_1 = np.sqrt(
            lamda ** 2 + 1j * w * mu_0 * (1 + chi[nlay - 2]) * sig[nlay - 2]
        )
        const0 = mu_0 * (1 + chi[nlay - 2]) / (mu_0 * (1 + chi[nlay - 1]) * utemp_1)

        dj0Mtemp00 = 0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
        dj0Mtemp10 = -0.5 * (const0)
        dj0Mtemp01 = -0.5 * (const0) * np.exp(-2.0 * utemp_1 * h_1)
        dj0Mtemp11 = 0.5 * (const0)

        dJ_10Mtemp00 = dj0Mtemp00
        dJ_10Mtemp10 = dj0Mtemp10
        dJ_10Mtemp01 = dj0Mtemp01
        dJ_10Mtemp11 = dj0Mtemp11

        dJ00.append(dudsig * dJ_10Mtemp00)
        dJ10.append(dudsig * dJ_10Mtemp10)
        dJ01.append(dudsig * dJ_10Mtemp01)
        dJ11.append(dudsig * dJ_10Mtemp11)

        for i in range(nlay):

            dJ0sum00 = np.zeros(lamda.size, dtype=complex)
            dJ0sum10 = np.zeros(lamda.size, dtype=complex)
            dJ0sum01 = np.zeros(lamda.size, dtype=complex)
            dJ0sum11 = np.zeros(lamda.size, dtype=complex)

            dJ1sum00 = np.zeros(lamda.size, dtype=complex)
            dJ1sum10 = np.zeros(lamda.size, dtype=complex)
            dJ1sum01 = np.zeros(lamda.size, dtype=complex)
            dJ1sum11 = np.zeros(lamda.size, dtype=complex)

            if i == 0:

                for j in range(nlay - 2):

                    if j == 0:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    else:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            elif (i > 0) & (i < nlay - 1):

                dJ0sum00 = M00[0]
                dJ0sum10 = M10[0]
                dJ0sum01 = M01[0]
                dJ0sum11 = M11[0]

                for j in range(nlay - 2):

                    if j == i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                        )

                    elif j < i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 1],
                            M10[j + 1],
                            M01[j + 1],
                            M11[j + 1],
                        )

                    elif j > i - 1:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 2],
                            M10[j + 2],
                            M01[j + 2],
                            M11[j + 2],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            elif i == nlay - 1:

                dJ0sum00 = M00[0]
                dJ0sum10 = M10[0]
                dJ0sum01 = M01[0]
                dJ0sum11 = M11[0]

                for j in range(nlay - 1):

                    if j < nlay - 2:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            M00[j + 1],
                            M10[j + 1],
                            M01[j + 1],
                            M11[j + 1],
                        )

                    elif j == nlay - 2:

                        dJ1sum00, dJ1sum10, dJ1sum01, dJ1sum11 = matmul(
                            dJ0sum00,
                            dJ0sum10,
                            dJ0sum01,
                            dJ0sum11,
                            dJ00[i],
                            dJ10[i],
                            dJ01[i],
                            dJ11[i],
                        )

                    dJ0sum00 = dJ1sum00
                    dJ0sum10 = dJ1sum10
                    dJ0sum01 = dJ1sum01
                    dJ0sum11 = dJ1sum11

            drTE[i, :] = dJ1sum01 / M1sum11 - M1sum01 / (M1sum11 ** 2) * dJ1sum11

    return drTE
    # return rTE, drTE



from . import Utils
from . import Props
import numpy as np
import scipy.sparse as sp
import uuid
import gc


class BaseRx:
    """SimPEG Receiver Object"""

    locs = None  #: Locations (nRx x nDim)

    knownRxTypes = (
        None  #: Set this to a list of strings to ensure that srcType is known
    )

    projGLoc = "CC"  #: Projection grid location, default is CC

    storeProjections = True  #: Store calls to getP (organized by mesh)

    def __init__(self, locs, rxType, **kwargs):
        self.uid = str(uuid.uuid4())
        self.locs = locs
        self.rxType = rxType
        self._Ps = {}
        Utils.setKwargs(self, **kwargs)

    @property
    def rxType(self):
        """Receiver Type"""
        return getattr(self, "_rxType", None)

    @rxType.setter
    def rxType(self, value):
        known = self.knownRxTypes
        if known is not None:
            assert value in known, "rxType must be in ['{!s}']".format(
                "', '".join(known)
            )
        self._rxType = value

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs.shape[0]

    def getP(self, mesh, projGLoc=None):
        """
            Returns the projection matrices as a
            list for all components collected by
            the receivers.

            .. note::

                Projection matrices are stored as a dictionary listed by meshes.
        """
        if mesh in self._Ps:
            return self._Ps[mesh]

        if projGLoc is None:
            projGLoc = self.projGLoc

        P = mesh.getInterpolationMat(self.locs, projGLoc)
        if self.storeProjections:
            self._Ps[mesh] = P
        return P


class BaseTimeRx(BaseRx):
    """SimPEG Receiver Object"""

    times = None  #: Times when the receivers were active.
    projTLoc = "N"

    def __init__(self, locs, times, rxType, **kwargs):
        self.times = times
        BaseRx.__init__(self, locs, rxType, **kwargs)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs.shape[0] * len(self.times)

    def getSpatialP(self, mesh):
        """
            Returns the spatial projection matrix.

            .. note::

                This is not stored in memory, but is created on demand.
        """
        return mesh.getInterpolationMat(self.locs, self.projGLoc)

    def getTimeP(self, timeMesh):
        """
            Returns the time projection matrix.

            .. note::

                This is not stored in memory, but is created on demand.
        """
        return timeMesh.getInterpolationMat(self.times, self.projTLoc)

    def getP(self, mesh, timeMesh):
        """
            Returns the projection matrices as a
            list for all components collected by
            the receivers.

            .. note::

                Projection matrices are stored as a dictionary (mesh, timeMesh)
                if storeProjections is True
        """
        if (mesh, timeMesh) in self._Ps:
            return self._Ps[(mesh, timeMesh)]

        Ps = self.getSpatialP(mesh)
        Pt = self.getTimeP(timeMesh)
        P = sp.kron(Pt, Ps)

        if self.storeProjections:
            self._Ps[(mesh, timeMesh)] = P

        return P


class BaseSrc(Props.BaseSimPEG):
    """SimPEG Source Object"""

    loc = None  #: Location [x,y,z]

    rxList = None  #: SimPEG Receiver List
    rxPair = BaseRx

    def __init__(self, rxList, **kwargs):
        assert type(rxList) is list, "rxList must be a list"
        for rx in rxList:
            assert isinstance(rx, self.rxPair), "rxList must be a {!s}".format(
                self.rxPair.__name__
            )
        assert len(set(rxList)) == len(rxList), "The rxList must be unique"
        self.uid = str(uuid.uuid4())
        self.rxList = rxList
        Utils.setKwargs(self, **kwargs)

    @property
    def nD(self):
        """Number of data"""
        return self.vnD.sum()

    @property
    def vnD(self):
        """Vector number of data"""
        return np.array([rx.nD for rx in self.rxList])


class Data:
    """Fancy data storage by Src and Rx"""

    def __init__(self, survey, v=None):
        self.uid = str(uuid.uuid4())
        self.survey = survey
        self._dataDict = {}
        for src in self.survey.srcList:
            self._dataDict[src] = {}
        if v is not None:
            self.fromvec(v)

    def _ensureCorrectKey(self, key):
        if type(key) is tuple:
            if len(key) != 2:
                raise KeyError("Key must be [Src, Rx]")
            if key[0] not in self.survey.srcList:
                raise KeyError("Src Key must be a source in the survey.")
            if key[1] not in key[0].rxList:
                raise KeyError("Rx Key must be a receiver for the source.")
            return key
        elif isinstance(key, self.survey.srcPair):
            if key not in self.survey.srcList:
                raise KeyError("Key must be a source in the survey.")
            return key, None
        else:
            raise KeyError("Key must be [Src] or [Src,Rx]")

    def __setitem__(self, key, value):
        src, rx = self._ensureCorrectKey(key)
        assert rx is not None, "set data using [Src, Rx]"
        assert isinstance(value, np.ndarray), "value must by ndarray"
        assert (
            value.size == rx.nD
        ), "value must have the same number of data as the source."
        self._dataDict[src][rx] = Utils.mkvc(value)

    def __getitem__(self, key):
        src, rx = self._ensureCorrectKey(key)
        if rx is not None:
            if rx not in self._dataDict[src]:
                raise Exception("Data for receiver has not yet been set.")
            return self._dataDict[src][rx]

        return np.concatenate([self[src, rx] for rx in src.rxList])

    def tovec(self):
        return np.concatenate([self[src] for src in self.survey.srcList])

    def fromvec(self, v):
        v = Utils.mkvc(v)
        assert v.size == self.survey.nD, "v must have the correct number of data."
        indBot, indTop = 0, 0
        for src in self.survey.srcList:
            for rx in src.rxList:
                indTop += rx.nD
                self[src, rx] = v[indBot:indTop]
                indBot += rx.nD


class BaseSurvey:
    """Survey holds the observed data, and the standard deviations."""

    std = None  #: Estimated Standard Deviations
    eps = None  #: Estimated Noise Floor
    dobs = None  #: Observed data
    dtrue = None  #: True data, if data is synthetic
    mtrue = None  #: True model, if data is synthetic

    counter = None  #: A SimPEG.Utils.Counter object

    def __init__(self, **kwargs):
        Utils.setKwargs(self, **kwargs)

    srcPair = BaseSrc  #: Source Pair

    @property
    def srcList(self):
        """Source List"""
        return getattr(self, "_srcList", None)

    @srcList.setter
    def srcList(self, value):
        assert type(value) is list, "srcList must be a list"
        assert np.all(
            [isinstance(src, self.srcPair) for src in value]
        ), f"All sources must be instances of {self.srcPair.__name__!s}"
        assert len(set(value)) == len(value), "The srcList must be unique"
        self._srcList = value
        self._sourceOrder = dict()
        [
            self._sourceOrder.setdefault(src.uid, ii)
            for ii, src in enumerate(self._srcList)
        ]

    def getSourceIndex(self, sources):
        if type(sources) is not list:
            sources = [sources]
        for src in sources:
            if getattr(src, "uid", None) is None:
                raise KeyError("Source does not have a uid: {!s}".format(str(src)))
        inds = list(map(lambda src: self._sourceOrder.get(src.uid, None), sources))
        if None in inds:
            raise KeyError(
                "Some of the sources specified are not in this survey. "
                "{!s}".format(str(inds))
            )
        return inds

    @property
    def prob(self):
        """
        The geophysical problem that explains this survey, use::

            survey.pair(prob)
        """
        return getattr(self, "_prob", None)

    @property
    def mesh(self):
        """Mesh of the paired problem."""
        if self.ispaired:
            return self.prob.mesh
        raise Exception("Pair survey to a problem to access the problems mesh.")

    def pair(self, p):
        """Bind a problem to this survey instance using pointers"""
        assert hasattr(p, "surveyPair"), "Problem must have an attribute 'surveyPair'."
        assert isinstance(self, p.surveyPair), (
            "Problem requires survey object must be an instance of a {!s} "
            "class.".format(p.surveyPair.__name__)
        )
        if p.ispaired:
            raise Exception(
                "The problem object is already paired to a survey. " "Use prob.unpair()"
            )
        elif self.ispaired:
            raise Exception(
                "The survey object is already paired to a problem. "
                "Use survey.unpair()"
            )
        self._prob = p
        p._survey = self

    def unpair(self):
        """Unbind a problem from this survey instance"""
        if not self.ispaired:
            return
        self.prob._survey = None
        self._prob = None

    @property
    def ispaired(self):
        return self.prob is not None

    @property
    def nD(self):
        """Number of data"""
        return self.vnD.sum()

    @property
    def vnD(self):
        """Vector number of data"""
        return np.array([src.nD for src in self.srcList])

    @property
    def nSrc(self):
        """Number of Sources"""
        return len(self.srcList)

    @Utils.count
    @Utils.requires("prob")
    def dpred(self, m=None, f=None):
        """dpred(m, f=None)

            Create the projected data from a model.
            The fields, f, (if provided) will be used for the predicted data
            instead of recalculating the fields (which may be expensive!).

            .. math::

                d_\\text{pred} = P(f(m))

            Where P is a projection of the fields onto the data space.
        """
        if f is None:
            f = self.prob.fields(m)
        return self.eval(f)

    @Utils.count
    def eval(self, f):
        """eval(f)

            This function projects the fields onto the data space.

            .. math::

                d_\\text{pred} = \\mathbf{P} f(m)
        """
        raise NotImplementedError("eval is not yet implemented.")

    @Utils.count
    def evalDeriv(self, f):
        """evalDeriv(f)

            This function s the derivative of projects the fields onto the data space.

            .. math::

                \\frac{\\partial d_\\text{pred}}{\\partial u} = \\mathbf{P}
        """
        raise NotImplementedError("eval is not yet implemented.")

    @Utils.count
    def residual(self, m, f=None):
        """residual(m, f=None)

            :param numpy.array m: geophysical model
            :param numpy.array f: fields
            :rtype: numpy.array
            :return: data residual

            The data residual:

            .. math::

                \\mu_\\text{data} = \\mathbf{d}_\\text{pred} - \\mathbf{d}_\\text{obs}

        """
        return self.dpred(m, f=f) - self.dobs

    @property
    def isSynthetic(self):
        "Check if the data is synthetic."
        return self.mtrue is not None

    def makeSyntheticData(self, m, std=0.05, f=None, force=False):
        """
            Make synthetic data given a model, and a standard deviation.

            :param numpy.array m: geophysical model
            :param numpy.array std: standard deviation
            :param numpy.array u: fields for the given model (if pre-calculated)
            :param bool force: force overwriting of dobs

        """
        if getattr(self, "dobs", None) is not None and not force:
            raise Exception(
                "Survey already has dobs. You can use force=True to override "
                "this exception."
            )
        self.mtrue = m
        self.dtrue = self.dpred(m, f=f)
        noise = std * abs(self.dtrue) * np.random.randn(*self.dtrue.shape)
        self.dobs = self.dtrue + noise
        self.std = self.dobs * 0 + std
        return self.dobs


class LinearSurvey(BaseSurvey):
    def eval(self, f):
        return f

    @property
    def nD(self):
        return self.prob.G.shape[0]



from scipy.constants import G
from SimPEG.Utils import mkvc
import numpy as np


def GravSphereFreeSpace(x, y, z, R, xc, yc, zc, rho):
    """
        Computing the induced response of magnetic sphere in free-space.
        >> Input
        x, y, z:   Observation locations
        R:     radius of the sphere
        xc, yc, zc: Location of the sphere
        rho: Density of sphere

        By convention, z-component is positive downward for a
        positive density anomaly

    """
    if ~np.size(x) == np.size(y) == np.size(z):
        print("Specify same size of x, y, z")
        return

    unit_conv = 1e8  # Unit conversion from SI to (mgal*g/cc)
    x = mkvc(x)
    y = mkvc(y)
    z = mkvc(z)

    nobs = len(x)

    M = R ** 3.0 * 4.0 / 3.0 * np.pi * rho

    rx = x - xc
    ry = y - yc
    rz = zc - z

    rvec = np.c_[rx, ry, rz]
    r = np.sqrt((rx) ** 2 + (ry) ** 2 + (rz) ** 2)

    g = -G * (1.0 / r ** 2) * M * unit_conv

    gx = g * (rx / r)
    gy = g * (ry / r)
    gz = g * (rz / r)

    return gx, gy, gz
